// Copyright (c) 2010-2017 Trevor Blight ////
//
// system header file for p5c generated code
//

#ifndef __P5C_H__
#define __P5C_H__  100

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <stdbool.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <setjmp.h>
#include <unistd.h>

// translate pascal names to equivalent c name
static inline double arctan( double x ) { return atan(x); }
static inline double ln( double x ) { return log(x); }

// file handling
typedef struct {
    FILE *f;
    char *name;
    int flags;
    unsigned char buffer;
} text;

struct _PcleanupInf {
     int level;
     void (*cuFunc)(void);
     struct _PcleanupInf *link;
};
extern struct _PcleanupInf *_Phead;

static void  __attribute__ ((__noreturn__))
_Pfatal( char *const s, const int l )
{
fprintf( stderr, "fatal error at line %d: %s\n", l, s);
exit(EXIT_FAILURE);
} // _Pfatal()


#if defined MEMDEBUG
struct _Pmem
{
    struct _Pmem *left;
    struct _Pmem *right;
    int line;
    bool f;
    void * addr;
};
struct _Pmem* _Proot=NULL;

static void _Padd_mem( struct _Pmem **const m, void *const a, const int l )
{
struct _Pmem *p = *m;
    if( p==NULL ) {
        p = (struct _Pmem*)malloc( sizeof(struct _Pmem) );
        if( p != NULL ) {
            *m = p;
            p->left = NULL;
            p->right = NULL;
            p->line = l;
            p->f = 0;
            p->addr = a;
        }
        else
            _Pfatal("out of memory", l);
    }
    else if( a < p->addr )
        _Padd_mem(&p->left, a, l );
    else if( a > p->addr )
        _Padd_mem(&p->right, a, l );
    else if( p->f != 0 ) {
        p->line = l;
        p->f = 0;
    }
    else
        _Pfatal("memory already allocated", l);
} // _Padd_mem()

static void _Pfree_mem( struct _Pmem *const p, void *const a, const int l )
{
char msg[50];
    if( p == NULL ) _Pfatal( "attempt to dispose nil pointer", l);
    if( a > p->addr )
       _Pfree_mem( p->right, a, l );
    else if( a < p->addr )
       _Pfree_mem( p->left, a, l );
    else if( p->f == 0 ) {
       free(a);
       p->line = l;
       p->f = !0;
    }
    else {
       sprintf(msg, "memory was already disposed at line %d", p->line );
       _Pfatal( msg, l);
    }
} // _Pfree_mem()

static void _Puse_mem( struct _Pmem *const p, void *const a, const int l )
{
    if( p == NULL ) _Pfatal( "attempt to use invalid pointer", l);
    if( a > p->addr )
       _Puse_mem( p->right, a, l );
    else if( a < p->addr )
       _Puse_mem( p->left, a, l );
    else if( p->f != 0 ) {
       char msg[40];
       sprintf(msg, "using pointer disposed at line %d", p->line );
       _Pfatal( msg, l);
    }
} // _Puse_mem()

static void _Preport_mem( struct _Pmem *const p )
{
    if( p == NULL ) return;
    _Preport_mem( p->left );
    if( p->f == 0 ) {
        printf( "new memory at line %i was not disposed\n", p->line );
    }
    _Preport_mem( p->right );
}
#endif // defined MEMDEBUG

#endif //__P5C_H__
