program Toy{Prolog} (input, output, mc, si, so);

{ (C) COPYRIGHT 1983  --  Feliks Kluzniak, Stanislaw Szpakowicz
			  Institute of Informatics, Warsaw University   }

label 1, 2;  { error halt & almost-fatal error recovery only }


const
	maxint       =  32767;
	eos             =       0 ;     { end-of-string character ord }
	eol             =       10;     { end-of-line character ord }
	maxchar         =       127;    { maximum character ord }

	{  table limits  }
	ctlow   =  0;   cthigh  =  4000;        { character table }
	atsize  =  1009;                         { atom table size: a prime }
	atlow   =  2;   athigh  =  1010;         { athigh=atlow-1+atsize }
	ttlow   =  0;   tthigh  =  800;         { trail table }
	btlow   =  0;   bthigh  =  400;         { backtrack point table }
	btbelow = -1;                   { = btlow - 1, for initialisation }
	ftlow   =  0;   fthigh  =  800;         { frame table }
	ftbelow = -1;                   { = ftlow - 1, for initialisation }
	mtlow   =  2;   mthigh  =  16000;                { main table }

	{  at  }
	noarity =  -1;          { used when an atom is only a name handle }
	emptyatom       =  cthigh;    {name value marking unused atom entries
					can't be a name - no room for eos }

	{  mt ,  see also var declarations  }
	stacksize  =  5000;     { size of var & copy stack area }
	int        =  1;        { marker for integer objects, = varlim+1 }
	varlim     =  0;        { fields holding  <= varlim are vars,
				   only dumvarx holds varlim            }
	freevar    =  -1;       { filler for free vars in stack area }
	offoff     =  2;        { variable prot with offset k is
				   represented as  - ( k + offoff )     }
	minstdint =  -1;        { integers in this range are given stan- }
	maxstdint =  10;        {     dard prototypes to save memory     }

	{  various  }
	maxvarnumb  =  mthigh;          { max number of different vars
						in a clause - 1         }
	maxargnumb  =  mthigh;          { max number of args in a term }
	maxsyspars =  5;                { max number of parameters in a
						system routine          }
	car     =  1;           { dot's argument number ( sequence head ) }
	cdr     =  2;           { _______________ ,, _____________ tail   }

	{ i / o }
	maxlinesize = 80;

	{ term i/o - ops }
	maxpriority   = 1201;            { priority used for brackets etc. }
	maxoppriority = 1200;            { max for ops}
	commapriority = 1000;            { standard op ','/2 }
	minpriority   = 0;
	nopriority  = minpriority;  { Prolog-10 convention: higher is weaker }
	ophashsize  = 17;           { size of declaration hash table }

	{ term i/o - scanner }

	{ character class encodings, see array ccclass }
	{note that <= csmalletter is letter and <= cunderscore is alphameric}
	cbigletter = 'a';   csmalletter = 'b';   cdigit = 'c';
	cunderscore = 'd';
	clbracket = 'e';   crbracket = 'f';
	csolochar = 'g';   csymch = 'h';
	cidquote = 'i';   cstringquote = 'j';   cpercent = 'k';
	cperiod = 'l';   csign = 'm';   cbar = 'n';   cblank = 'o';
	cother = 'p';

	{ term i/o - parser }

	{ opop table entries, coded as characters }
	conflict   = 'A';   { ambiguity    = 'B'; ---UNUSED!--- }
	shiftifarg = 'C';   shiftifnoarg = 'D';
	redifarg   = 'E';   redifnoarg   = 'F';

	{ eqpr table entries, coded as characters }
	badmatch       = '0';   shiftbar       = '1';
	closeidcumargs = '2';   closenilorlist = '3';
	closebarlist   = '4';   closeround     = '5';
	closecurly     = '6';


type
	{  index ranges  }
	ctx     =  ctlow .. cthigh;     { character table index }
	atx     =  atlow .. athigh;     { atom table index }
	ttx     =  ttlow .. tthigh;     { trail table index }
	btx     =  btlow .. bthigh;     { backtrack point table index }
	ftx     =  ftlow .. fthigh;     { frame table index }
	ftblwx  =  ftbelow..fthigh;	{ addition for TROOVERLAY }
	mtx     =  mtlow .. mthigh;     { main table index }

	stackx  =  mtx; { i.e.mtlow .. stackhigh; stack area index in mt }
	protx   =  mtx; { i.e.protlow .. mthigh;  prototype area index in mt}

	{  sundry ranges  }
	varnumb  =  0 .. maxvarnumb;    { number of vars / var offset }
	argnumb  =  noarity .. maxargnumb;      { number of arguments }
	nsysparam =  0 .. maxsyspars;   { number of system routine params }
	classofchar     =   cbigletter..cother;
	classesofchar   =   set of classofchar;

	{ standard atoms ( indexing table std ) :
			';'/2, ','/2, 'lcurly rcurly'/1  'call'/1, 'tag'/1,
			'[]'/0, '.'/2, 'error'/1, '#'/0,  'user'/0   }
	stdatomid  =  ( atmsemcol,  atmcomma,  atmcurly,  atmcall,  atmtag,
			atmnil,  atmdot,  atmerror,  atmgo,  atmuser         );


	{  system routine identifiers  }
	sysroutid  =  ( idfail,  idtag,  idcall,
			idslash,
			idtagcut,   idtagfail,   idtagexit,  idancestor,
			idhalt,   idstatus,
			idop,   iddelop,
			idwrite,  idwriteq,  idread,
			iddisplay,  idrch,  idlastch,  idskipbl,  idwch,
			idecho,  idnoecho,  idsee,  idseeing,  idseen,
			idtell,  idtelling,  idtold,
			idordchr,  idsum,  idprod,  idinf,  idnameinf,
			idsmalletter,  idbigletter,  idletter,  iddigit,
			idalphanum,  idbracket,  idsolochar,  idsymch,
			ideqvar,  idvar,  idatom,  idinteger,  idnonvarint,
			idfunctor,  idarg,  idpname,  idpnamei,
			idproc,  idproclimit,   idprocinit,
			idclause,  idretract,  idabolish,
			idassert,  idredefine,  idpredefined,  idprotect,
			idnonexistent,  idnononexistent,
			iddebug,  idnodebug,
			idtrue                  );

	{  error codes  }
	errid      =  ( ctovflw,  atovflw,  ttovflw,  btovflw,
			ftovflw,  stackovflw,  protovflw,
			syntax,  loadfile,  sysinit,   usereof,
			longfilename,  unifyerr,  badkernel    );



	{  objects  }

	procptr  =  ^ proc;

	proc  =         record          { a procedure descriptor }
			  next   :  procptr;    { what to do after failure }
			  nvars  :  varnumb;    { size of variable frame }
			  case systemroutine : boolean of
			    true  : ( sysid : sysroutid );
			    false : ( head, body : protx )
			end;

	atom  =         record  { unique functor/predicate representation }
			  name   :  ctx;        { printname string,
						   = emptyatom if not used }
			  arity  :  argnumb;
			  proc   :  procptr     { associated proc or nil }
			end;

	backpt  =       record          { backtrack-point (fail-point) }
			  frozenframe  :  ftx;          { where to resume }
			  resume       :  procptr;      { what to resume }
			  copylev      :  stackx;       { copy stack level }
			  traillev     :  ttx           { trail stack level}
			end;

	frame   =       record          { environment frame }
			  anc    :  ftx;        {to caller frame (but TRO)}
			  calls  :  mtx;        { call is 1st in this seq. }
			  vars   :  stackx      { variables of this frame }
			end;


	{ term i/o - ops }

	priority    =  minpriority .. maxpriority;
	oppriority  =  minpriority .. maxoppriority;
	opkind      =  ( noop, fx, fy, xfy, xfx, yfx, yf, xf );
	opkinds     =  set of opkind;
	ophashx     =  1 .. ophashsize;

	opptr       =  ^ opdecl;
	opdecl      =  record           { op declaration list element }
			 next     : opptr;     { in this hash chain }
			 name     : atx;       {name/0 used,as in op call}
			 priority : oppriority;
			 kinds    : opkinds    { at most 1 unary & 1 binary }
		       end;


	{ term i/o - scanner }

	{ noarg is a var, a string, an int or an id (not op) with no args.
	   idarg is an id (not op) with immediately following '('.
	   dot is a period followed by a layout character.                 }
	lexclass   =  ( noarglex,  oplex,  idarglex,
			lroundlex,  lcurlylex,  lsquarelex,  dotlex,
			barlex,  rsquarelex,  rcurlylex,  rroundlex  );

	token  =  record
		    prior : priority;
		    case class : lexclass of
		      noarglex   :  ( term : mtx );
		      oplex      :  ( opname : atx;  kinds : opkinds );
		      idarglex   :  ( idname : atx;  idarity : argnumb );
		      lroundlex, lcurlylex  :  ();
		      lsquarelex :  ( listlength : integer );
		      barlex, rsquarelex,
		      rcurlylex, rroundlex  :  ()
		  end;


	{ term i/o - parser }

	synterrid  =  ( bug, badchar, bigint, opbad, ambig, nodelim,
			noarg, nomatch, badbar, roundempty, curlyempty,
			unclosed, empty                                     );

	parselptr = ^ parsel;
	parsel  =  record    { parse stack element }
		     prev : parselptr;    {previous: the stack is a LIFO list}
		     lex  : token
		   end;

	opopaction  =  conflict .. redifnoarg;
	eqpraction  =  badmatch .. closecurly;


var
	{ mt address constants ( alas, no compile-time expressions ) }
	stackhigh,      { mtlow-1+stacksize, top of var & copy stack area }
	protlow,        { stackhigh+1, bottom of prototype area }
	dumvarx,        { protlow, the unique dummy variable prototype }
	nilprotx,       { dumvarx+1, a prototype of '[]'/0 }
	minchprotx,     { nilprotx+1, low end of unique char prot area }
	maxchprotx,     { minchprotx+maxchar, high end of char prot area }
	stdintprotx     { maxchprotx+1, low end of std int prot area }
		: mtx;

	{** variables **}

	mc, si, so : text; { sysfile,seefile,tellfile }
	seeing, telling :  boolean;     { terminal/file i/o flags }
	seeobj, tellobj :  mtx;         { terms with current file names }
	sysloading  :  boolean;         { on when loading system file }
	outlinesize, solinesize : 0 .. maxlinesize;
	echo     : boolean;     { input echo switch }

	ct      :  array [ ctx ] of char;       { character string table }
	ctop ,                          { first free above names }
	cbot ,                          { first free below temp var names }
	builtname ,                     { first char of constructed name }
	builtvar  :  ctx;               { last (lowest) char of currently
					   processed variable name        }

	at      :  array [ atx ] of atom;       { atom table : hashed }
	freeatoms : integer;                    { >= 1, 1 kept as sentinel }

	tt      :  array [ ttx ] of stackx;     { trail table }
	ttop    :  ttx;                         { first free }

	bt      :  array [ btx ] of backpt;     { bactrack-point table }
	topb    :  btbelow .. bthigh;   { last used = current backpt }

	ft      :  array [ ftx ] of frame;      { environment frame table }
	topf    :  ftblwx;   { last used = current frame }

	mt      :  array [ mtx ] of integer;    { main memory table }
	vtop ,                          { first free above variable stack }
	csbot ,                         { first free below term copy stack }
	gtop ,                  { first free above ground prot area }
	ngbot   :  mtx;         { first free below non-ground prot area }
	gprotect, ngprotect : protx;  { clauses with heads below or above
					     these can't be redefined         }

	{ storage utilisation :  low level markers, for status only }
	tminfree ,   bminfree ,   fminfree ,
	stackminfree ,   cminfree ,   pminfree  :  integer ;

	{ control state registers ( see also topf and topb ) }
	ancf    :  ftblwx;         { = ft [ topf ]. anc }
	topenv ,                { = ft [ topf ]. vars, current clause's
						variable environment    }
	ancenv :  stackx;       { = ft [ ancf ]. vars, current call's
						variable environment    }
	ccall    :  mtx;        { the current call }
	cproc    :  procptr;    { the current proc }

	frozenvars ,            { = ft [ bt [ topb  ]. frozenframe ]. vars }
	frozenheap  :  stackx;  { = bt [ topb ]. copylev }

	debug    :  boolean;    { wallpaper trace switch }

	{ standard system routine interface }
	syserror  :  boolean;   { turned on for bad system routine calls }
	spar  :  array [ 1 .. maxsyspars ] of mtx; {parameters (in ancenv)}
	sparv :  array [ 1 .. maxsyspars ] of integer; {values of int spars}

	{ standard atom pointers ( constants ) }
	std  :  array [ stdatomid ] of atx;

	{ pointers to standard objects in prototype area ( constants) }
	errorcallseq    :  protx;               { '.'('error'(:0), '[]') }
	userprot        :  protx;               { prototype of 'user'/0 }

	{ assert etc. }
	lastasserted    : atx;  {head atom of most recently asserted clause}
	redefining      : boolean;      { redefinition switch, cf.redefine }
	vardictbot      : stackx;       { first location of temporary
					   variable dictionary above
					   topmost variable frame - used
					   only by storeclause & helpers &
					   term i/o                         }
	procindx        : integer;      { used by genproc & helpers }
	shownonexistent : boolean;      { show calls to empty procs ? }

	{ reader & high i/o }
	cch      : char;        { current input character }

	{ term i/o - ops }
	opatoms : array [fx..xf] of atx;    { atoms of op decl patterns }
	opht    : array [ophashx] of opptr; {opdecl chains hashed on pname}
	commaopname,  baropname  :  atx;      { ','/0 & '|'/0, protected }
	binaryops, unaryops, rightops,
	leftops, prefixops, infixops,
	postfixops                     :  opkinds;

	{ term i/o - scanner }

	inlex   :   token;                 { scanner-parser interface }
	cc : classofchar;                  { class of current char (cch) }

	{ character class mapping the classes are encoded as characters,
	   as this is the only way to obtain a packed table with some
	   Pascal implementations.                                          }
	ccclass   :       array [ char ] of classofchar;


	{ term i/o - parser }

	justreduced : boolean;      { was the last action a reduction ? }
	toplex      : token;        { top element of parse stack }
	cellar      : parselptr;    { other elements kept in LIFO list }
	freeparsels : parselptr;    { don't rely on dispose }
	savegtop    : protx;        { gtop is used for argument stack }

	{ opop determines legal action when both tokens are ops of equal
	   priorities. entries coded as chars to get it packed.          }
	opop        :  array[ fx..xf, fx..xf ] of opopaction;

	{ eqpr is similar to opop : it is used for maxpriority
	   tokens ( idarg, brackets & bar )                     }
	eqpr  :  array[ idarglex..barlex, barlex..rroundlex ] of eqpraction;



	{ main program variables }
	terminate  :   boolean;         { flag set by system routine halt }
	goalstmnt  :  protx;            { current initial goal statement }
	ngoalvars  :  varnumb;          { and the number of its variables }
	dummy      :  boolean;          { needed to fill parameter of doread }



{---------  e r r o r s   ---------}

procedure putline;
forward; {-----}

procedure halt;
{ this might be implementation-dependent }
begin   writeln ( output );   writeln (output, ' **** toyprolog aborted ****');
	goto 1
end;

procedure status;
const field = 8;
type tablename = array [ 1..9 ] of char;
	procedure line ( caption : tablename;  size, free, unused : integer );
	begin writeln ( output,
			  caption, size : field, free : field, unused : field )
	end;
begin   writeln ( output );
	writeln ( output, ' Toy Prolog, version 2.1 -  IIUW Warszawa 1983' );
	writeln ( output, '  table    size    free   unused' );
	line ( 'character', cthigh, cthigh - ctop, cminfree );
	line ( 'atom     ', atsize, freeatoms, freeatoms );
	line ( 'prototype', mthigh - protlow, ngbot - gtop, pminfree );
	line ( 'trail    ', tthigh, tthigh - ttop, tminfree );
	line ( 'backpt   ', bthigh, bthigh - topb, bminfree );
	line ( 'frame    ', fthigh, fthigh - topf, fminfree );
	line ( 'var/copy ', stackhigh - mtlow, csbot - vtop, stackminfree );
	outlinesize:= 0
end;

procedure errorr ( id : errid );
begin   writeln ( output );
	write ( output, ' ++++ error : ' );
	case id of
	  ctovflw       :  write ( output, 'character' );
	  atovflw       :  write ( output, 'atom' );
	  ttovflw       :  write ( output, 'trail' );
	  btovflw       :  write ( output, 'backtrack-point' );
	  ftovflw       :  write ( output, 'frame' );
	  stackovflw    :  write ( output, 'variable/copy' );
	  protovflw     :  write ( output, 'prototype' );
	  syntax        :  write ( output, 'bad system program syntax' );
	  loadfile      :  write ( output, 'bad system kernel file' );
	  sysinit       :  write ( output, 'interpreter initialisation error' );
	  usereof       :  write ( output, 'user eof ' );
	  longfilename  :  write ( output, 'file name too long' );
	  unifyerr      :  write ( output, 'error in unify, interpreter bad' );
	  badkernel     :  write ( output, 'error in kernel file' )
	end{case};
	if id <= protovflw then begin
		write ( output, ' table overflow' );
		putline;   status
	end
	else putline;
	if id in [ctovflw,protovflw,loadfile,sysinit,usereof,badkernel] then
		halt
	else goto 2
end;

(***** procedure check ( n : integer;  b : boolean ); *****)
       { debugging help :  abort if b false }
(***** begin   if not b then begin
 *****                 writeln ( output );
 *****                 writeln (output, ' +-+-+-+-+ check failure ', n );
 *****                 halt
 *****         end
 ***** end;
 *****)





{---------  l o w   i n p u t / o u t p u t  ---------}

	{ file opening/closing is implementation-dependent.
	   see also procedure loadsyskernel ( initialisation ). }

procedure closefile ( forinput : boolean );
{ close mc,si or so, switching to user terminal.
   forinput signals whether the current input or the current output file
   is to be closed. no effect if the file is not open.                   }
begin   if forinput then begin
		if seeing then begin
			seeing:= false;   seeobj := userprot
		end
		else
		if sysloading then begin
			sysloading := false;   seeobj := userprot
		end
	end
	else    if telling then begin
			telling:= false;   tellobj := userprot
		end
end {closefile};

procedure openfile ( name : ctx;  forinput : boolean );
{ open the file (si or so according to 2nd parameter)
   whose name is given by a string in character table.
   a file is opened rewound. if a previous file is open, it is closed.  }
const  ln = 35; { for RSX-11 }
var   nm : array [ 1..ln ] of char;   k : 1..ln;
begin   k:= 1;
	while ( k <> ln ) and ( ct [ name ] <> chr ( eos ) ) do begin
		nm [ k ] :=  ct [ name ];   name:= name + 1;   k:= k + 1
	end;
	if ct [ name ] <> chr ( eos ) then errorr ( longfilename );
	for k:= k to ln do nm [ k ] := ' ';
	closefile ( forinput );             { only 1 file per stream }
	if forinput then begin
		reset ( si, nm );   seeing:= true
	end
	else begin
		rewrite ( so, nm );   telling:= true;
		solinesize:= 0
	end
end {openfile};

procedure putline;
{ start a new output line }
begin   if telling then begin
		 writeln ( so );   solinesize:= 0
	end
	else begin
		writeln ( output );   outlinesize:= 0
	end
end;

procedure getline;
{ start a new input line }
var   keep : boolean;
begin   if sysloading then readln ( mc )
	else
	if seeing then readln ( si )
	else readln ( input );
	if echo then begin
		 keep:= telling;   telling:= false;
		 putline;    telling:= keep
	end
end;

procedure putch ( c : char );
{ put a character onto the current output file }
{ characters are counted to avoid line overflow }
begin   if telling then begin
		if solinesize >= maxlinesize then putline;
		write( so, c );   solinesize:= solinesize +1
	end
	else begin
		if outlinesize >= maxlinesize then putline;
		write( output, c );   outlinesize:= outlinesize + 1
	end
end;

procedure getch ( var c : char );
{ get a character from the current input file }
var   keep : boolean;
begin   if sysloading then read ( mc, c )
	else
	if seeing then read ( si, c )
	else read ( input, c );
	if echo then begin
		keep:= telling;   telling:= false;
		putch ( c );   telling:= keep
	end
end;

procedure putint ( i : integer );
{ output an integer without blanks }
var   divisor : integer;
begin   if i < 0 then begin
		putch ( '-' );   i:= - i
	end;
	divisor:= 10000;                        {**16 bit version**}
	while ( i div divisor = 0 ) and ( divisor <> 1 ) do
						divisor:= divisor div 10;
	repeat  putch ( chr ( i div divisor + ord ( '0' ) ) );
		i:= i mod divisor;   divisor:= divisor div 10
	until divisor = 0
end;

function endofline  : boolean;
{ eoln on the current input file ? }
begin   if sysloading then endofline:= eoln ( mc )
	else
	if seeing then endofline:= eoln ( si )
	else endofline:= eoln ( input )
end;

function endoffile  : boolean;
{ eof on the current input file ? }
begin   if sysloading then endoffile:= eof ( mc )
	else
	if seeing then endoffile:= eof ( si )
	else endoffile:= eof ( input )
end;





{---------  d i c t i o n a r y  ---------}

	{ string representation is known to initprot, gterm & nameexplode }
	{ scanner also does some hocus pocus with ct }


procedure cspace;
var   left : integer;
begin   left:= cbot - ctop - 1;
	if left < 0 then errorr ( ctovflw );
	if left < cminfree then cminfree:= left
end;

function compstrings ( s1, s2 : ctx;  equality : boolean )  : boolean;
{ equality --> s1 = s2 ?  #  not equality --> s1 < s2 ?  }
begin   while ( ct[s1] =  ct[s2])  and  ( ct[s2] <>  chr(eos) ) do begin
		s1:= s1 + 1;   s2:= s2 + 1
	end;
	if equality then compstrings:= ct[s1] = ct[s2]  { both are eos ? }
	else compstrings:= ct[s1] < ct[s2]      { note that eos = 0 }
end;

function charlast ( string : ctx )  : ctx;
{ locate the last character (except eos) of this string }
begin   while  ct [ string ] <> chr ( eos )  do string:= string + 1;
	charlast:= string - 1           {correct if lowest string not empty}
end;

procedure newatom ( where : atx;  printname : ctx;  narity : argnumb );
{ create a new atom ( at least one empty atom needed as sentinel ) }
begin   if freeatoms = 1 then errorr ( atovflw );
	with at [ where ] do begin
		name:= printname;   arity:= narity;   proc:= nil
	end;
	freeatoms:= freeatoms - 1
end;

function findname ( firstch, lastch : ctx )  : atx;
{ find the first occurence of an atom with name in ct firstch
   through lastch. create one ( with noarity ) if none .        }
var   a : atx;
begin   { a simple hash on first and last char }
	{ ***linear rehash used : don't forget findatom when improving*** }
	a:= ( ord(ct[firstch])*256 + ord(ct[lastch]) ) mod atsize  + atlow;
	{ rehash loop stopped by empty sentinel ( freeatoms >= 1 ) }
	while ( at [ a ]. name <> emptyatom ) and
			not compstrings ( at [ a ]. name, firstch, true )  do
		if a <> athigh then a:= a + 1
		else a:= atlow;
	if at [ a ]. name = emptyatom then newatom ( a, firstch, noarity );
	findname:= a
end;

function findatom ( nameatom : atx;  narity : argnumb )  : atx;
{ given the first atom with name, find or create one with name and arity }
var   a : atx;   thename : ctx;
begin   if at [ nameatom ]. arity = noarity then begin {just a name handle:}
		at [ nameatom ].arity:= narity;        { so use it !       }
		findatom:= nameatom
	end
	else begin      { linear rehash (there is an empty atom sentinel) }
		thename:= at [ nameatom ]. name;   a:= nameatom;
		while ( at [ a ]. name <> emptyatom ) and
			 ( (at[a].name<>thename) or (at[a].arity<>narity) ) do
			if a <> athigh then a:= a + 1
			else a:= atlow;
		if at [a]. name = emptyatom then newatom(a, thename, narity);
		findatom:= a
	end
end;

procedure buildname ( c : char );
{ add a character to the currently-constructed name }
begin   cspace;
	ct [ ctop ]:= c;   ctop:= ctop + 1
end;

function wrapname  : atx;
{ the constructed name is complete :  return the first atom with this name }
var   nm : atx;
begin   buildname ( chr ( eos ) );              { terminate the string }
	nm:= findname ( builtname, ctop - 2 );
	if at [ nm ]. name <> builtname then    { copy already exists    }
		ctop:= builtname                { so destroy the new one }
	else builtname:= ctop;                  { prepare for next name }
	wrapname:= nm
end;

function samename ( atom : atx;  narity : argnumb )  : atx;
{ return the first atom ( in the appropriate hash chain ) with
   the same name as atom but with arity ( which may be different )      }
begin   with at [ atom ] do
	samename:= findatom ( findname ( name , charlast ( name ) ), narity )
end;

procedure putname ( nm : atx );
{ output the print name of this atom }
var   c : ctx;
begin   c:= at[nm].name;
	while ct [ c ] <> chr ( eos ) do begin
		putch ( ct [ c ] );   c:= c + 1
	end
end;

procedure putqname ( nm : atx );
{ output this name in quotes }
var   c : ctx;
begin   putch ( '''' );   c:= at[ nm ].name;
	while ct[ c ] <> chr( eos ) do begin
		putch ( ct[ c ] );
		if ccclass [ ct[ c ] ] = cidquote then putch ( ct[ c ] );
		c:= c + 1
	end;
	putch ( '''' )
end;

function needsquoting ( nm : atx ) : boolean;
{ should this name be quoted to be readable for Prolog ? }
var   c : ctx;   class : classofchar;   ok : boolean;
begin   c:= at[nm].name;
	class:= ccclass[ct[c]];   ok:= true;
	if class = csolochar then ok:= ct[c + 1] = chr ( eos )     {1char?}
	else
	if ( class = cperiod ) and ( ct[ c + 1 ] = chr ( eos ) ) then  {dot}
		ok:= false
	else
	if class in [csymch,csign,cperiod] then
		while ok and ( ct[c + 1] <> chr ( eos ) ) do begin
			c:= c + 1;
			ok:= ccclass[ct[c]] in [csymch,csign,cperiod]
		end
	else
	if class = csmalletter then                                  {ident?}
		while ok and ( ct[c + 1] <> chr ( eos ) ) do begin
			c:= c + 1;
			ok:= ccclass[ct[c]] in
				  [csmalletter,cbigletter,cdigit,cunderscore]
		end
	else ok:= false;
	needsquoting:= not ok
end;





{---------  t e r m   r e p r e s e n t a t i o n  ---------}

	{ pt stands for a prototype or a term instance }
	{ note: chars have unique prots, cf.newprot & charprot,
		 so have integers in the most frequently used range }


procedure protspace ( size : integer );
{ enough space in prototype area ? }
var   left : integer;
begin   left:= ngbot - gtop - size;
	if left < size then errorr ( protovflw );
	if left < pminfree then pminfree:= left
end;

(***** function isnongroundprot ( pt : mtx )  : boolean; *****)
       { is it a non-ground prototype ( a prot with variables ) ? }
       { expanded in varadr and terminst }
(***** begin   isnongroundprot:=  pt > ngbot
 ***** end;
 *****)

function isvar ( pt : mtx )  : boolean;
{ a variable ? }
begin   isvar:=  mt [ pt ] <= varlim
end;

function isndvar ( pt : mtx )  : boolean;
{ a non-dummy variable ? }
{ expanded in unify, terminst & argument }
begin   isndvar:= mt [ pt ] < varlim
end;

(***** function isframevar ( t : mtx )  : boolean; *****)
       { a variable environment stack var ? }
       { expanded in bind2vars & terminst }
(***** begin   isframevar:=  t < vtop
 ***** end;
 *****)

function isint ( pt : mtx )  : boolean;
{ an integer ? }
{ expanded in unify }
begin   isint:=  mt [ pt ] = int
end;

function intval ( pt : mtx )  : integer;
{ an integer :  return its value }
{ expanded in unify }
begin intval:=  mt [ pt + 1 ]
end;

function is_normal ( pt : mtx )  : boolean;
{ not a variable or integer ?   note that int = varlim + 1 }
{ expanded in unify }
begin   is_normal:=  mt [ pt ]  >  int
end;

function isargless ( pt : mtx )  : boolean;
{ a non-variable, non-integer prot/term without arguments ? }
begin   if mt [ pt ] <= int then isargless:= false
	else isargless :=   at [ mt [ pt ] ] . arity  =  0
end;

function ischar ( pt : mtx;  var c : char )  : boolean;
{ is it a character term ?  if so, what is the character ? }
var   val : integer;
begin   val:= pt - minchprotx;
	if ( val < 0 ) or ( val > maxchar ) then ischar:= false
	else begin
		c:= chr ( val );   ischar:= true
	end
end;

function charprot ( c : char )  : protx;
{ return this character's prototype }
begin   charprot:= minchprotx + ord ( c )
end;

function getarity ( pt : mtx )  : argnumb;
{ return the narity of pt's functor }
{ expanded in unify and terminst }
begin   getarity:=  at [ mt [ pt ] ] . arity
end;

function getarg ( pt : mtx;  n : argnumb )  : mtx;
{ return pt's n'th argument (for prots only prot arg returned) }
{ expanded in argument }
begin   { check(1, is_normal( pt ) and ( n>= 0 ) and ( n <= getarity( pt ))); }
	if mt [ pt + n ] <= varlim then getarg:= pt + n
	else getarg:= mt [ pt + n ]
end;

function isstdint ( val : integer;  var prot : protx )  : boolean;
{ if val is a standard integer, return its unique prototype }
begin   if ( val < minstdint ) or ( val > maxstdint ) then isstdint:= false
	else begin
		prot:= stdintprotx + ( val - minstdint ) * 2;
		isstdint:= true
	end
end;

procedure mkint ( val : integer );
{ build an integer prototype on top of ground area }
begin   protspace ( 2 );
	mt [ gtop ] := int;   mt [ gtop + 1 ] := val;   gtop:= gtop + 2
end;

function newintprot ( val : integer )  : protx;
{ return an integer prototype (ground) }
var   p : protx;
begin   if isstdint ( val, p ) then newintprot:= p
	else begin
		newintprot:= gtop;   mkint ( val )
	end
end;

function initprot ( atom : atx )  : protx;
{ create a prototype whose functor is described by atom.
   note that it will need argument filling and wrapping.
   guards uniqueness of character prototypes.            }
var   size : integer;   achar : boolean;
begin   with at [ atom ] do begin
		size:= 1 + arity;               { 1 for the functor }
		if size <> 1 then achar:= false
		else achar:= ct [ name + 1 ] = chr ( eos );
		if achar then initprot:= charprot ( ct [ name ] )
	end;
	if not achar then begin
		protspace ( size );   ngbot:= ngbot - size;
		mt [ ngbot + 1 ] := atom;               { functor filled }
		initprot:= ngbot + 1
	end
end;

procedure newpvararg ( p : protx;  n : argnumb;  v : varnumb );
{ insert this variable offset as the n'th argument of a constructed prot }
begin   mt [ p + n ] :=  - ( v + offoff )
end;

procedure newparg ( p : protx;  n : argnumb;  arg : integer );
{insert the n'th (non-var or dummy var) argument into constructed prototype}
begin   { check (2, arg >= protlow ); }
	mt [ p + n ] := arg
end;

function wrapprot ( p : protx )  : protx;
{ p is a recently constructed prototype. return its final address
   after attempting to move it to the ground prot area. the manner of
   construction ensures that p can't be ground if it is not directly above
   ngbot ( or already in the ground area - applies to chars & std ints ).   }
var   arity, k : argnumb;   ground : boolean;
begin   if p <> ngbot + 1 then wrapprot:= p
	else begin                      { see if all args are ground }
		arity:= getarity ( p );   k:= 0;   ground:= true;
		while ( k <> arity ) and ground do begin
			k:= k + 1;   ground :=  getarg ( p, k )  <  gtop
		end;
		if not ground then wrapprot:= p
		else begin                      { shift to ground area }
			wrapprot:= gtop;
			for k:= 0 to arity do begin     {0 for the functor}
				ngbot:= ngbot + 1;
				mt [ gtop ] := mt [ ngbot ];
				gtop:= gtop + 1
			end
		end {ground}
	end {ground?}
end {wrapprot};





{---------  t e r m   i n s t a n c e s   ---------}

	{ pt stands for a copy stack term instance or
	   for a prototype in variable environment env }


procedure stackspace ( size : integer );
var   left : integer;
begin   left:= csbot - ( vtop + size );
	if left < 0 then errorr ( stackovflw );
	if left < stackminfree then stackminfree:= left
end;

procedure pushvars ( n : varnumb );
{ push n free variables onto the variable environment stack }
begin   if n <> 0 then begin    { test just for efficiency }
		stackspace ( n );
		for n:= n downto 1 do begin
			mt [ vtop ] :=  freevar;   vtop:= vtop + 1
		end
	end
end;

function heapint ( i : integer )  : mtx;
{ create an integer object on the copy stack or return std int prot }
var   p : protx;
begin   if isstdint ( i, p ) then heapint:= p
	else begin
		stackspace ( 2 );
		mt [ csbot ] :=  i;   mt [ csbot - 1 ] :=  int;
		heapint:= csbot - 1;   csbot:= csbot - 2
	end
end;

function heapdot ( head, tail : mtx )  : stackx;
{ create an instance of  .(head, tail)  }
begin   stackspace ( 3 );
	mt [ csbot ] :=  tail;   mt [ csbot - 1 ] :=  head;
	mt [ csbot - 2 ] :=  std [atmdot];   heapdot:= csbot - 2;
	csbot:= csbot - 3
end;

function gterm ( atom : atx )  : stackx;
{ create a most general term instance for this atom, i.e. a term with
   different variables for arguments, but guard char uniqueness.           }
var   thisarity : argnumb;   thisname : ctx;
begin   with at [ atom ] do begin
		thisarity:= arity;   thisname:= name
	end;
	if ( thisarity = 0 ) and ( ct [ thisname + 1 ] = chr( eos ) ) then
					gterm:= charprot ( ct [ thisname ] )
	else begin
		stackspace ( 1 + thisarity );   { 1 for the functor }
		for thisarity:= thisarity downto 1 do begin
			mt [ csbot ] :=  freevar;   csbot:= csbot - 1
		end;
		mt [ csbot ] :=  atom;   gterm:= csbot;   csbot:= csbot - 1
	end
end;

procedure bindv ( v : stackx;  t : mtx );
{ bind variable instance v to term t, pushing a trail if necessary }
begin   { check (3, isndvar ( v ) and ( mt [ v ] = freevar ) ); }
	mt [ v ]:=  - t;
	if ( v > frozenheap ) or ( v < frozenvars ) then begin
		if ttop = tthigh then errorr ( ttovflw );
		tt [ ttop ] := v;   ttop:= ttop + 1;
		if tminfree > tthigh - ttop then tminfree:= tthigh - ttop
	end
end;

procedure undo ( tlev : ttx );
{ bring trail level down to tlev and undo the bindings referenced by
   popped trail entries.                                                }
begin   while ttop > tlev do begin
		ttop:= ttop - 1;   mt [ tt [ ttop ] ] :=  freevar
	end
end;

procedure bind2vars ( v1, v2 : stackx );
{ v1 and v2 are free variable instances. if they are of the same kind,
   bind the younger to the older, otherwise bind the frame variable to
   the copy stack ( heap ) variable.
   ** v < vtop is expanded isframevar                                   }
begin   { check (4, isndvar ( v1 ) and isndvar ( v2 ) ); }
	if v1 = v2 then         { no action - must remain free }
	else
	if v1 < vtop then                                       {v1 frame}
		if v2 < vtop then                               {v2 frame}
			if v1 < v2 then bindv ( v2, v1 )         {v1 older}
			else bindv ( v1, v2 )                    {v2 older}
		else bindv ( v1, v2)                     {v1 frame, v2 heap}
	else
	if v2 < vtop then bindv ( v2, v1 )               {v2 frame, v1 heap}
	else
	if v1 < v2 then bindv ( v1, v2 )                         {v2 older}
	else bindv ( v2, v1 )                                    {v1 older}
end;

function varadr ( v : mtx;  env : stackx )  : stackx;
{ v is a variable prototype or instance. if a prototype,
   find its instance in variable environment env. **isnongroundprot expanded }
begin   { check (5, isndvar ( v ) ); }
	if not ( v > ngbot ) then varadr:= v
	else varadr:= env + ( - mt [ v ] - offoff )
end;

function argument ( pt : mtx;  env : stackx;  n : argnumb )  : mtx;
{ pt is a term instance or a prototype in variable environment env,
   and is neither a variable nor an integer. return its dereferenced
   n'th argument. instead of a variable prototype return
   a frame variable cell index. ** getarg expanded here.                }
var   field, arg : mtx;
begin   field:= pt + n;
	if mt [field] > varlim then arg:= mt [field]            { non-var }
	{ mt [ field ] = varlim impossible, as only dumvarx holds varlim }
	else begin      { non-dummy var, dereference binding if any }
		arg:= varadr ( field, env );
		{note that nonvars or free vars will stop the loop }
		while mt [ arg ] < freevar do arg:= - mt [ arg ]
	end;
	argument:= arg
end;

function terminst ( pt : mtx;  env : stackx )  : mtx;
{ if pt is a term instance ( a ground prototype or a copy-stack object ),
   return its address. if it is a non-ground prototype use env as
   the proper variable environment and return the address of a new copy
   in the copy stack. free frame variables are bound to their copies.     }
{ ***isnongroundprot, getarity and isndvar expanded }
var   arity, k : argnumb;   copy : stackx;   arg : mtx;
begin   if not ( pt > ngbot ) then terminst:= pt     { not isnongroundprot }
	else    { the following branch for stand-alone frame variables }
	if mt[pt] < varlim then begin    { isndvar: a new heap variable }
		stackspace ( 1 );
		mt [ csbot ] := freevar;   copy:= csbot;   csbot:= csbot - 1;
		bindv ( varadr ( pt, env ), copy );   terminst:= copy
	end
	else begin
		arity:= at[ mt[ pt ] ] .arity;
		stackspace ( 1 + arity );       { 1 for the functor }
		csbot:= csbot - ( 1 + arity );   copy:= csbot + 1;
		mt [ copy ] := mt [ pt ];               { functor }
		for k:= 1 to arity do begin             { arguments }
			arg:= argument ( pt, env, k );
			if arg < vtop then begin  {framevar,evidently free}
				bindv ( arg, copy + k );
				mt [ copy + k ] :=  freevar
			end
			else mt [ copy + k ] := - terminst ( arg, env )
			     {ie.args always linked as if through vars,
			       because terminst can return a variable.  }
		end;
		terminst:= copy
	end
end;

function unify ( c : mtx;  cenv : stackx;  h : mtx;  henv : stackx )
								: boolean;
{ c and h are terms ( possibly prototypes in variable environments cenv and
   henv ), but neither is a non-dummy variable prototype ( cf. argument ).
   try to unify, but don't undo on failure.                                 }
var   success : boolean;   arity, n : argnumb;   mth, mtc : integer;
begin   { check(6, not ( isndvar( c ) and isnongroundprot( c ) ) and
			not ( isndvar( h ) and isnongroundprot( h ) ) ); }
	success:= true;         { let's be optimistic for a change }
	mth:= mt [ h ];   mtc:= mt [ c ];       { more wieldy & efficient }
	{ first branch if same prots and environment or same terms and
	   environment doesn't matter (i.e. not non-ground prots),
	   i.e. an attempt to unify a term with itself : no action required }
	if ( h = c ) and ( ( henv = cenv ) or ( h <= ngbot ) )  then
	else
	if ( mth > int ) and ( mtc > int ) then         {both are "normal" }
		if mth <> mtc then success:= false      {different functors}
		else begin                              {try the arguments }
			n:= 0;   arity:= at [ mth ]. arity;
			while ( n <> arity ) and success do begin
				n:= n + 1;
				success:= unify( argument( c, cenv, n ), cenv,
						 argument( h, henv, n ), henv)
			end
		end
	else                                    { at least one int or var }
	if ( h = dumvarx ) or ( c = dumvarx ) then    {no action}
	else
	if mth = freevar then                                    {isndvar(h)}
		if mtc = freevar then bind2vars ( h, c )         {isndvar(c)}
		else bindv ( h, terminst ( c, cenv ) )
	else
	if mtc = freevar then bindv ( c, terminst ( h, henv ) )   {isndvar(c)}
	else
	if mth = int then                                       {isint(h)}
		if mtc <> int then success:= false      { not isint( c ) }
		else success:= mt [ h + 1 ] = mt [ c + 1 ]      {intvals}
	else
	if mtc = int then success:= false
	else errorr ( unifyerr );
	unify:= success
end;

procedure display ( pt : mtx;  env : stackx );
{ output pt in standard form }
var   arity, k : argnumb;
begin   if isvar ( pt ) then begin
		putch ( '_' );
		if pt <> dumvarx then putint ( varadr ( pt, env ) )
	end
	else
	if isint ( pt ) then putint ( intval ( pt ) )
	else begin
		putname ( mt[pt] );   arity:= getarity ( pt );
		if arity <> 0 then begin
			putch ( '(' );   putch ( ' ' );
			for k:= 1 to arity do begin
				display ( argument ( pt, env, k ), env );
				if k <> arity then putch ( ',' );
				putch ( ' ' )
			end;
			putch ( ')' )
		end {arity<>0}
	end
end {display};

function iscallsequence ( pt : mtx;  env : stackx ) : boolean;
{ is this term a dotted list without integer or variable elements ? }
var   is : ( no, dontknow, yes );
begin   is:= dontknow;
	repeat  if not ( is_normal ( pt ) ) then is:= no
		else
		if mt [ pt ] = std [atmnil] then is:= yes
		else
		if mt [ pt ] <> std [atmdot] then is:= no
		else begin
			if not is_normal( argument(pt, env, car)) then is:= no;
			pt:= argument ( pt, env, cdr )          { tail }
		end
	until is <> dontknow;
	iscallsequence:=  is = yes
end;





{---------  c o n t r o l  ---------}

	{ note that call sequence prototypes have non-variable elements.
	   thus mt [ seq + car ] and mt [ seq + cdr ] access the first
	   call in a sequence and the sequence tail respectively.         }


procedure dowrite ( term : mtx;  q : boolean );
forward; {- - - - - - - - - - - - - - - - - }

procedure sysroutcall ( id : sysroutid;  var success, stop : boolean );
forward;  {- - - - - - - - - - - - - - - - - - - - - - - - - - - - -}


function isnonunitclause ( proc : procptr )  : boolean;
{ is proc a clause, and does it have any calls ? }
begin   if proc = nil then isnonunitclause:= false
	else with proc ^  do
		if systemroutine then isnonunitclause:= false
		else isnonunitclause:= mt [ body ] <> std [atmnil]
end;

procedure pushbackpt ( proc : procptr );
{ push a backtrack point freezing the current frame for resumption of proc }
begin   if topb = bthigh then errorr ( btovflw );
	topb:= topb + 1;
	if bminfree > bthigh - topb then bminfree:= bthigh - topb;
	with bt [ topb ] do begin
		frozenframe:= topf;   resume:= proc;
		copylev:= csbot;   traillev:= ttop
	end;
	frozenvars:= topenv;   frozenheap:= csbot
end;

procedure popbackpt;
{ topb <> btlow. return to the previous backtrack point }
begin   { check (7, topb <> btlow ); }
	topb:= topb - 1;
	with bt [ topb ] do begin
		frozenheap:= copylev;   frozenvars:= ft [ frozenframe ]. vars
	end
end;

procedure pushframe ( ancestor : ftx;  callsequence : protx );
{ create a new (current) environment frame with no variables (as yet) }
begin   if topf = fthigh then errorr ( ftovflw );
	topf:= topf + 1;
	if fminfree > fthigh - topf then fminfree:= fthigh - topf;
	with ft [ topf ] do begin
		anc:= ancestor;   calls:= callsequence;   vars:= vtop
	end;
	topenv:= vtop;   ancf:= ancestor;   ancenv:= ft [ ancf ]. vars
end;

procedure backtrack ( var stop : boolean );
{ set stop if no backtrack points left; otherwise restore the state frozen
   by the current backpt and set cproc to the clause to be attempted next   }
begin   if topb = btlow then stop:= true
	else begin
		csbot:= frozenheap;   vtop:= frozenvars;
		with bt [ topb ] do begin
			undo ( traillev );
			if frozenframe <> topf then begin  { not shallow }
				topf:= frozenframe;   topenv:= ft [topf].vars;
				ancf:= ft [topf].anc; ancenv:= ft [ancf].vars;
				{ get ccall, skipping call & tag :
				   no error possible, we've been here }
				ccall:= mt [ ft [ topf ]. calls + car ];
				while ( mt [ ccall ] = std [atmcall] ) or
				      ( mt [ ccall ] = std [atmtag] ) do
					ccall:= argument( ccall, ancenv, 1 )
			end;
			cproc:= resume ^. next;
			if cproc = nil then popbackpt
			else
			if cproc ^. next = nil then popbackpt
			else resume:= cproc
		end {with}
	end
end {backtrack};

procedure exitt ( var stop : boolean );
{ starting from the most recently activated procedure ( system routine or
   unit clause ) exit until a pending call is found. set stop if none.     }
var   f : ftx;   nextcseq : protx;   found : boolean;
begin   found:= false;   f:= topf;   stop:= false;
	repeat  { tentative exit until nonterminated frame or stack bottom }
		nextcseq:= mt [ ft [ f ]. calls + cdr ];
		if mt [ nextcseq ] = std [atmdot] then found:= true
		else begin                              { no more calls }
			f:= ft [ f ]. anc;   stop :=  f = ftlow
		end
	until found or stop;
	if found then begin     { finalise exit and get next proc }
		ancf:= ft [ f ]. anc;   ancenv:= ft [ ancf ]. vars;
		if f > bt [ topb ]. frozenframe then topf:= f
		else begin              { frozen : must activate a copy }
			topf:= bt [ topb ]. frozenframe + 1;
			with ft [ topf ] do begin
				anc:= ancf;   vars:= frozenvars
			end
		end;
		with ft [ topf ] do begin       { advance in frame }
			calls:= nextcseq;   topenv:= vars
		end;
		vtop:= topenv   { discard exited nonfrozen var env frames }
	end
end {exitt};

function candotro ( frame : ftx )  : boolean;
{ can we apply tail recursion optimisation (tro) to this frame, i.e.
   destroy it after shifting its variables over those of its ancestor ?
   we can if its call is last in its clause ( and is not an outermost
   tag ) and the ancestor is not frozen.                                }
begin   with ft [ frame ] do
		if anc <= bt [ topb ]. frozenframe then candotro:= false
		else candotro:= ( mt [ mt [calls + cdr] ] = std [atmnil] ) and
				( mt [ mt [ calls + car ] ] <> std [atmtag] )
end;

procedure trooverlay ( var youngframe : ftblwx;  var youngenv : stackx;
				var oldframe : ftblwx;  var oldenv : stackx );
{ apply tail recursion optimisation :  overlay old frame's variable
   environment (oldenv) with that of youngframe (youngenv), then
   let youngframe reference oldframe and oldframe - oldframe's ancestor.
   youngenv must be the youngest variable environment on the stack.
   both frames must not be frozen (vars  not referenced from trail)     }
var   low, high : stackx;   shiftby, v : integer;
begin   if oldenv <> youngenv then begin  {shift after reference adjustment}
		high:= youngenv;   shiftby:= youngenv - oldenv;
		while high < vtop do begin  {adj. refs to shifted/overlaid}
			v:=  - mt [ high ];     { value of shifted var }
			if ( v >= oldenv ) and ( v < youngenv ) then begin
				{ref to overlaid. if overlaid free, reverse
				  pointers so that all vars bound to it will
				  be rebound to high; otherwise follow the
				  overlaid ref & repeat adjustment loop for
				  this high to deal with new mt[high].      }
				if mt [ v ] = freevar then begin
					mt [ high ] :=  freevar;
					mt [ v ] :=  - high;
					high:= high + 1
				end
				else mt [ high ] := mt [ v ]
			end {to overlaid}
			else begin   {if ref. to shifted, adjust to new adr}
				if ( v >= youngenv ) and ( v < vtop ) then
					mt [ high ] :=  - ( v - shiftby );
				high:= high + 1
			end
		end {adjustment loop};
		low:= oldenv;
		for high:= youngenv to vtop - 1 do begin        { shift }
			mt [ low ] :=  mt [ high ];   low:= low + 1
		end;
		vtop:= low
	end {tricky dealings};
	youngenv:= oldenv;   youngframe:= oldframe;
	oldframe:= ft [ youngframe ]. anc;   oldenv:= ft [ oldframe ]. vars
end {trooverlay};

procedure enter;
{ build a frame for the first call in cproc }
begin   { check (8, ( cproc <> nil ) and not cproc^.systemroutine ); }
	if candotro ( topf ) then trooverlay ( topf, topenv, ancf, ancenv );
	pushframe ( topf, cproc ^. body )
end;

procedure errsetup;
{ invoked upon detection of an erroneous system routine call ( so topf is not
   frozen ! ). top frame is associated with a variable bound to the bad call,
   then  a call to 'error'/1 is set up.                                     }
begin   { check (9, topf > bt [ topb ]. frozenframe ); }
	topenv:= vtop;   ft [ topf ]. vars:= topenv;   pushvars ( 1 );
	bindv ( topenv,   terminst ( mt [ ft [topf].calls + car ], ancenv ) );
	pushframe ( topf, errorcallseq );   ccall:= mt [ errorcallseq + car ]
end;

procedure preparestep;
{ load ccall, stripping off tag & call, then load cproc and
   push a backtrack-point if needed.                            }
begin   ccall:= mt [ ft [ topf ]. calls + car ];
	while ( mt [ ccall ] = std [atmcall] ) or
	      ( mt [ ccall ] = std [atmtag] ) do begin
		ccall:= argument ( ccall, ancenv, 1 );
		if not is_normal ( ccall ) then errsetup      {affects ccall}
	end;
	cproc:= at [ mt [ ccall ] ]. proc;
	if cproc <> nil then
		if cproc ^. next <> nil then pushbackpt ( cproc )
end {preparestep};

procedure onestep ( var success, stop : boolean );
{ perform a single resolution step.
   on entry success is false immediately after a backtrack, true otherwise. }
	procedure showcall;
	{ show the call in wallpaper trace }
	var  keep : boolean;
	begin   keep:= telling;   telling:= false;
		display ( ccall, ancenv );  telling:= keep
	end;
	procedure showresult;
	{ show the call's result in wallpaper trace }
	var  keep : boolean;
	begin   keep:= telling;   telling:= false;
		if success then writeln ( output, '*+*' )
		else writeln ( output, '*-*' );
		putline;   telling:= keep
	end;
	procedure shownoproc;
	{ output a message for a call to a nonexistent procedure }
	var  keep : boolean;
	begin   keep:= telling;   telling:= false;
		putline;
		write(output,'---failing call to non existent procedure:');
		putline;   putname ( at[mt[ccall]].name );
		putch ( '/' );
		putint( getarity ( ccall ) );   putline;
		telling:= keep
	end;

begin   if success then preparestep;    { otherwise backtrack did }
	if debug then showcall;
	if cproc = nil then begin             {empty proc called}
		success:= false;
		if shownonexistent then shownoproc
	end
	else with cproc ^  do begin
		pushvars ( nvars );
		if systemroutine then sysroutcall ( sysid, success, stop )
		else success:= unify( ccall, ancenv, cproc ^. head, topenv )
	end {with};
	if debug then showresult
end {onestep};

procedure startresolution ( goalsequence : protx;  ngoalvars : varnumb );
{ initialise stacks in preparation for this goal's execution. the lowermost
   frame is associated with goal's variables, the lowermost backtrack point
   freezes it and acts as a sentinel. both have some dummy entries.         }
begin   ttop:= ttlow;   vtop:= mtlow;   csbot:= stackhigh;
	topf:= ftbelow;   topb:= btbelow;
	pushframe ( ftlow, nilprotx );                  { dummy call }
	pushbackpt ( nil );
	pushvars ( ngoalvars );   pushframe ( topf, goalsequence )
end;

procedure resolve ( goalsequence : protx;  nvarsingoalseq : varnumb );
{ execute this ( initial ) goal sequence }
var   success, stop : boolean;
begin   startresolution ( goalsequence, nvarsingoalseq );
	stop:= false;   success:= true;  { false only on failure }
	repeat
		onestep ( success, stop );
		if not stop then
			if syserror then begin
				syserror:= false;   success:= true;
				errsetup
			end
			else
			if not success then backtrack ( stop )
			else
			if isnonunitclause ( cproc ) then enter
			else exitt ( stop )
	until stop
end {resolve};




{---------  c o n t r o l   d i s r u p t i o n   e t c .  ---------}

	{ note that only explicit, outermost tags are detected.
	   eg. in "call ( tag [.])"  tag is ignored :
	   tags are tricky enough as it is.
	   cf. also candotro, backtrack and preparestep in control. }

procedure purgetrail ( low : ttx );
{ remove unnecessary trail entries at and above low , either after
   a successful unifyordont call or after popping backtrack-points in
   a non-backtracking context ( unfreeze ). this is necessary, as
   unfrozen vars might be moved or destroyed - it also saves trail space. }
var   high : ttx;
begin   for high:= low to ttop - 1 do
		if (tt [high] < frozenvars) or (tt [high] > frozenheap)
								    then begin
			tt [ low ] :=  tt [ high ];   low:= low + 1
		end;
	ttop:= low
end;

function unifyordont ( c : mtx; cenv : stackx; h : mtx; henv : stackx )
								    : boolean;
{ same as unify, but undo effects on failure. this is done by ensuring
   that all bindings are trailed.
    -----> used also outside control disruption <-----                  }
var   success : boolean;   fv, fh : stackx;   tlev : ttx;
begin   fv:= frozenvars;   fh:= frozenheap;   tlev:= ttop;
	frozenvars:= vtop;   frozenheap:= csbot;
	success:= unify ( c, cenv, h, henv );
	frozenvars:= fv;   frozenheap:= fh;
	if success then purgetrail ( tlev )
	else undo ( tlev );
	unifyordont:= success
end;

procedure unfreeze ( frame : ftx );
{ remove as many backtrack points as necessary to unfreeze frame (>ftlow) }
var   changed : boolean;
begin   changed:= false;
	while frame <= bt [ topb ]. frozenframe do begin {sentinel at btlow}
		changed:= true;   popbackpt
	end;
	if changed then purgetrail ( bt [ topb ]. traillev )
end;

procedure cutcleanup;
{ destroy cut-off frames between cut or tagcut and its immediate ancestor. }
var   ancancf : ftblwx;   ancancenv : stackx;   owncalls : protx;
begin   if topf <> ancf + 1 then begin
		vtop:= ft[ ancf + 1 ] .vars;
		ft[ ancf + 1 ] :=  ft[ topf ];   topf:= ancf + 1;
		ft[ topf ] .vars := vtop
	end;
	if candotro ( ancf ) then begin         { tro for ancf }
		ancancf:= ft [ ancf ]. anc;
		ancancenv:= ft [ ancancf ]. vars;
		owncalls:= ft [ topf ]. calls;   {save continuation}
		trooverlay( ancf, ancenv, ancancf, ancancenv );
		topf:= ancf + 1;   {  args were just changed }
		with ft [ topf ] do begin       {rebuild own frame}
			calls:= owncalls;   anc:= ancf;
			vars:= vtop;   topenv:= vtop
		end
	end {tro}
end {cutcleanup};

procedure slash;
{ perform the "cut" operation :  unfreeze the nearest ancestor which
   is not a ';'(_,_) or ','(_,_) (after stripping off "tag" and "call").
   attempt a delayed tro for the immediate ancestor.                    }
var  f,ancancf : ftx;  ancancenv : stackx;  found : boolean;
     call : mtx;
begin   f:= ancf;   found:= false;
	while ( f <> ftlow ) and not found do begin
		call:= mt [ ft [ f ]. calls + car ];
		ancancf:= ft [ f ]. anc;   ancancenv:= ft [ ancancf ]. vars;
		while (mt[call] = std[atmtag]) or (mt[call] = std[atmcall]) do
				       call:= argument ( call, ancancenv, 1 );
		if ( mt [ call ] <> std [ atmsemcol ] ) and
			( mt [ call ] <> std [ atmcomma ] ) then found:= true
		else f:= ancancf
	end;
	if found then begin
		unfreeze ( f );   cutcleanup
	end {found}
end {slash};

function gottagged ( var f : ftx;   pt : mtx;   env : stackx )  : boolean;
{ among the ancestors of f search for the nearest tagged call unifiable
   with pt in env. if found, unify and return true with f pointing at
   the tagged frame; if not, return false with f undefined.                }
var   call : protx;   callenv : stackx;   found : boolean;
begin   found:= false;
	while ( f <> ftlow ) and not found do
		with ft [ f ] do begin
			call:= mt [ calls + car ];   callenv:= ft [anc]. vars;
			if mt [ call ] = std [atmtag] then found:=
				unifyordont( argument(call, callenv, 1),
							callenv, pt, env );
			if not found then f:= anc
		end {with};
	gottagged:= found
end;

procedure ancestor ( var success : boolean );
{ try to unify sysrout arg with a tagged ancestor. }
var   frame : ftx;
begin   frame:= ancf;   success:= gottagged ( frame, spar [ 1 ], ancenv )
end;

procedure tagcut ( var success : boolean );
{ if a tagged ancestor unifies with the sysrout parameter, unfreeze it }
var   frame : ftx;
begin   frame:= ancf;
	if gottagged ( frame, spar [ 1 ], ancenv ) then begin
		unfreeze ( frame );   cutcleanup
	end
	else success:= false
end;


procedure tagfail ( var success : boolean );
{ if a tagged ancestor unifies with the sysrout parameter, cause
   the ancestor to fail immediately, otherwise simply fail.                }
begin   tagcut ( success );
	success:= false         { if ancestor unfrozen, it will fail now }
end;

procedure tagexit ( var success : boolean );
{ if a tagged ancestor unifies with the sysrout parameter, unify and
   immediately exit to the ancestor by resetting ancf and changing topf's
   call sequence to errorcallseq ( because it has an empty tail ! ),
   and cproc to nil. otherwise fail.                                      }
var   frame : ftx;
begin   frame:= ancf;
	if gottagged ( frame, spar [ 1 ], ancenv ) then
		with ft[ topf ] do begin
			calls:= errorcallseq;   cproc:= nil;
			anc:= frame;   ancf:= frame
		end
	else success:= false
end;




{---------  a r i t h m e t i c ,  c h a r s  ---------}

procedure bindpar ( v : nsysparam;  t : mtx );
{ v is the number of a variable sysrout parameter :  bind it to t.
   -----> used also outside arithmetic <-----                           }
begin   if spar [ v ] <> dumvarx then bindv ( varadr( spar [ v ], ancenv ), t )
end;

procedure arithans ( ansvar : nsysparam;  answer : integer );
{ sysrout parameter ansvar is a variable :  bind it to answer }
begin   bindpar ( ansvar, heapint ( answer ) )
end;

procedure arithpars ( k : nsysparam;  var code : integer );
{ k is the number of parameters in spar. for each integer parameter set
   a bit in the rightmost k bits of code. only integers and variables
   allowed:  syserror is set for others.                                    }
var   i : nsysparam;
begin   code:= 0;
	for i:= 1 to k do begin
		code:= code * 2;                { shift left }
		if isint ( spar [ i ] ) then code:= code + 1    { set bit }
		else
		if not isvar ( spar [ i ] ) then syserror:= true
	end
end {arithpars};

procedure sum ( var success : boolean );
{ a + b = x  or  a + x = c  or  x + b = c  or  a + b = c ? }
var   code : integer { 0..7 };
begin   arithpars ( 3, code );
	if not syserror then
		case code of
		0, 1, 2, 4 :  syserror:= true;  { not enough integers }
{011}         3 :  arithans ( 1, sparv [ 3 ] - sparv [ 2 ] );
{101}         5 :  arithans ( 2, sparv [ 3 ] - sparv [ 1 ] );
{110}         6 :  arithans ( 3, sparv [ 1 ] + sparv [ 2 ] );
{111}         7 :  success :=   sparv [ 1 ] + sparv [ 2 ] = sparv [ 3 ]
		end
end {sum};

procedure prod ( var success : boolean );
{ a*b+c = x  or  a*b+x = d  or  a*x+c = d or ( a*x+y = d  &  y < a ) or
   ( x*b+y = d  &  y < a )   or  x*b+c = d  or  a*b+c = d ?              }
var   code : integer { 0..15 };
	procedure divmod ( x, by : nsysparam );
	begin   if sparv [ by ] = 0  then syserror:= true
		else begin
			arithans ( 3, sparv [ 4 ] mod sparv [ by ] );
			arithans ( x, sparv [ 4 ] div sparv [ by ] )
		end
	end;
	procedure division ( x, by : nsysparam );
	begin   if sparv [ by ] = 0  then syserror:= true
		else begin
			if (sparv[ 4 ] - sparv[ 3 ]) mod sparv[ by ] <> 0
							then success:= false
			else arithans(x, (sparv[4] - sparv[3]) div sparv[by])
		end
	end;
begin   { prod }
	arithpars ( 4, code );
	if not syserror then
		case code of
		0,1,2,3,4,6,8,10,12 :  syserror:= true;  { ambiguous }
{0101}        5  :  divmod ( 1, 2 );
{0111}        7  :  division ( 1, 2 );
{1001}        9  :  divmod ( 2, 1 );
{1011}        11 :  division ( 2, 1 );
{1101}        13 :  arithans ( 3, sparv [ 4 ] - sparv [ 1 ] * sparv [ 2 ] );
{1110}        14 :  arithans ( 4, sparv [ 1 ] * sparv [ 2 ] + sparv [ 3 ] );
{1111}        15 :  success := sparv [1] * sparv [2] + sparv [3] = sparv [4]
		end
end {prod};

procedure inf ( var success : boolean );
{ a < b ? }
var   code : integer { 0..3 };
begin   arithpars ( 2, code );
	if not syserror then
		if code <> 3 then syserror:= true       { 2 ints needed }
{11}          else success :=  sparv [ 1 ]  <  sparv [ 2 ]
end;

procedure nameinf ( var success : boolean );
{ implement '@<' :  sysrout parameters are argless functors,
   check their lexicographic ordering.                          }
begin   if not isargless ( spar [ 1 ] ) or not ( isargless ( spar [ 2 ] ) )
							then syserror:= true
	else success:= compstrings( at [ mt [ spar [ 1 ] ] ]. name,
				    at [ mt [ spar [ 2 ] ] ]. name,  false )
end;

procedure ordchr ( var success : boolean );
{ chr ( a mod nch ) = x  or  x = ord ( b )  or  a mod nch = ord ( b ) ?
   nch is the number of characters, i.e. maxchar + 1 ( 128 for ASCII )   }
var   ch : char;   isch : boolean;
begin   isch:= ischar ( spar [ 2 ] , ch );      { ch defined if isch }
	if isint ( spar [ 1 ] ) then
		if isvar ( spar [ 2 ] ) then
			bindpar( 2, charprot( chr( sparv[1] mod (maxchar+1))))
		else
		if isch then success:= sparv[1] mod (maxchar+1) = ord ( ch )
		else syserror:= true
	else
	if isch and isvar ( spar [1] ) then arithans ( 1, ord ( ch ) )
	else syserror:= true
end {ordchr};

procedure ch1class ( class : classofchar;   var success : boolean );
{ if the sysrout parameter is not a character, raise syserror.
   otherwise check if it is of this class.                      }
var   c : char;
begin   if ischar ( spar [ 1 ], c ) then success:= ccclass [ c ] = class
	else success:= false
end;

procedure ch2class ( classes : classesofchar;  var success : boolean );
{ like chclass1, but a set of classes is given }
var   c : char;
begin   if ischar ( spar[ 1 ], c ) then success:= ccclass[ c ] in classes
	else success:= false
end;





{---------  s t r u c t u r e   a c c e s s  ---------}

function atomspar ( name, arity : nsysparam )  : atx;
{ extract an atom description from a sysrout call specifying name & arity.
   -----> used also outside structure access <-----                        }
begin   atomspar:= atlow;           { a dummy value in case of error }
	if not isargless( spar [name]) or not isint( spar [arity]) then
		syserror:= true
	else
	if sparv[ arity ] < 0 then syserror:= true
	else atomspar:= samename ( mt [spar [name]] , sparv [arity] )
end;

procedure eqvar ( var success : boolean );
{ are both sysrout parameters the same non-dummy variable ? }
begin   if not isndvar( spar[1]) or not isndvar( spar[2]) then success:= false
	else success :=  spar [ 1 ] = spar [ 2 ]
end;

procedure functor ( var success : boolean );
{ implement  functor( term, func, narity )  }
var   atom : atx;
begin   if isvar ( spar [ 1 ] ) then      { construct most general term }
		if isint ( spar [ 2 ] ) then
			if not isint ( spar [ 3 ] ) then syserror:= true
			else
			if sparv [ 3 ] <> 0 then syserror:= true
			else bindpar ( 1, spar [ 2 ] )
		else begin
			atom:= atomspar ( 2, 3 );
			if not syserror then bindpar ( 1, gterm ( atom ) )
		end
	else    { match term's functor with arity and func  }
	if isint ( spar [ 1 ] ) then
		if unify ( spar [ 2 ], ancenv, spar [ 1 ], vtop ) then
			success:= unify( spar[3], ancenv, heapint( 0 ), vtop )
		else success:= false
	else                                    { try narity, then func }
	if unify( spar[3], ancenv, heapint( getarity( spar[1] ) ), vtop ) then
		success:= unify( spar [ 2 ], ancenv,
				 gterm( samename( mt [ spar[1]] ,0 ) ), vtop )
	else success:= false
end {functor};

procedure arg ( var success : boolean );
{ implement  arg( number, term, argument )  }
begin   if not isint( spar[1]) or not is_normal( spar [2]) then syserror:= true
	else
	if (sparv[1]<1) or (sparv[1] > getarity(spar[2])) then success:= false
	else success:= unify ( spar [ 3 ], ancenv,
				argument (spar[2], ancenv, sparv[1]), ancenv )
end {arg};

function intexplode ( n : integer )  : mtx;
{ produce a list of n's digits ( in character form ) }
var   list : mtx;
begin   list:= nilprotx;
	repeat  list:= heapdot( charprot( chr( n mod 10 + ord('0'))), list );
		n:= n div 10
	until n = 0;
	intexplode:= list
end;

function nameexplode ( firstc : ctx )  : mtx;
{ produce a list of this printname's characters }
var   list : mtx;   lastc : ctx;
begin   list:= nilprotx;
	for lastc:= charlast ( firstc ) downto firstc do
		list:= heapdot ( charprot ( ct [ lastc ] ), list );
	nameexplode:= list
end;

function trydigseq ( list : mtx; env : stackx; var num : integer )  : boolean;
{ if list is a digit sequence ( in character form ), convert it to num,
 otherwise yield false. an empty list is  n o t  a digit sequence !     }
var   is : ( no, dontknow, yes );   dig : char;   d : integer;
      err, turn1 : boolean;
begin   is:= dontknow;   err:= false;   num:= 0;   turn1:= true;
	repeat  if not is_normal ( list ) then is:= no
		else
		if mt [ list ] = std [ atmnil ] then
			if turn1 then is:= no
			else is:= yes
		else
		if mt [ list ] <> std [ atmdot ] then is:= no
		else begin
			if not ischar( argument( list, env, car ), dig ) then
								       is:= no
			else
			if ( dig < '0' ) or ( dig > '9' ) then is:= no
			else
			if num > maxint div 10 then err:= true
			else begin
				num:= num * 10;   d:= ord( dig ) - ord( '0' );
				if num <> 0 then
					if ( maxint div num = 1 ) and
					   ( d > maxint mod num ) then
								   err:= true;
				if not err then num:= num + d
			end;
			list:= argument ( list, env, cdr );      {tail}
			turn1:= false
		end
	until ( is <> dontknow ) or err;
	trydigseq :=  is = yes
end;

function trycharseq ( list : mtx;  env : stackx;  var atom : atx )  : boolean;
{ if list is a character sequence, construct a name and return
   its atom ( null arity ).  otherwise yield false.
   an empty list  i s  a character sequence !                     }
var   is : ( no, dontknow, yes );   c : char;
begin   is:= dontknow;
	repeat  if not is_normal ( list ) then is:= no
		else
		if mt [ list ] = std [ atmnil ] then is:= yes
		else
		if mt [ list ] <> std [ atmdot ] then is:= no
		else begin
			if ischar ( argument ( list, env, car ), c ) then
							       buildname ( c )
			else is:= no;
			list:= argument ( list, env, cdr )      {tail}
		end
	until is <> dontknow;
	if is = no then ctop:= builtname        { restore dictionary state }
	else atom:= findatom ( wrapname, 0 );
	trycharseq :=  is = yes
end;

procedure pname ( var success : boolean );
{ implement  pname( functor, charlist )  }
var   list : mtx;   atom : atx;
begin   if not isvar ( spar [ 1 ] ) then                { explode }
		if not isargless ( spar [ 1 ] ) then syserror:= true
		else begin
			list:= nameexplode ( at [ mt [ spar [ 1 ] ] ] . name );
			success:= unify ( spar [ 2 ], ancenv, list, vtop )
		end
	else            { construct after checking for non-empty list }
	if not is_normal ( spar [ 2 ] ) then syserror:= true
	else
	if mt [ spar [ 2 ] ]  <>  std [atmdot] then syserror:= true
	else
	if trycharseq( spar[2], ancenv, atom ) then bindpar( 1, gterm( atom ))
	else syserror:= true
end {pname};

procedure pnamei ( var success : boolean );
{ implement  pnamei( integer, digitlist )  }
var   num : integer;
begin   if not isvar ( spar[ 1 ] ) then               { explode }
		if not isint ( spar[ 1 ] ) then syserror:= true
		else success:= unify ( spar[ 2 ], ancenv,
				       intexplode ( sparv [ 1 ] ), vtop )
	else
	if trydigseq ( spar [ 2 ], ancenv, num ) then
		bindpar ( 1, heapint ( num ) )
	else syserror:= true
end {pnamei};






{----------  c l a u s e / a s s e r t / r e t r a c t  ----------}

	 { assert uses free stack storage for temporary data }
	 { the area and the first few routines are also used
	    by the term writer                                }



function varconvert ( v : stackx )  : varnumb;
{ v is a variable instance. return its number in temporary
   variable dictionary, inserting it there if missing.          }
var   p : stackx;   found : boolean;
begin   p:= vardictbot;   found:= false;
	while ( p <> vtop ) and not found do
		if mt [ p ] =  -v then found:= true
		else p:= p + 1;
	if not found then begin                      { p = vtop }
		pushvars ( 1 );   mt [ p ]:=  - v
	end;
	varconvert:= p - vardictbot
end;

function storeterm ( pt : mtx;  env : stackx )  : protx;
{ used only by storeclause. pt is a non-variable term instance or
   prot in variable environment env. store it as a prototype, using
   temporary variable dictionary to renumber variables.                 }
var   p : protx;   k : argnumb;   arg : mtx;
begin   { check (10, not isndvar ( pt ) ); }
	if ( pt >= protlow ) and ( pt < gtop ) then storeterm:= pt {ground}
	else                            { heap integer still possible }
	if isint ( pt ) then storeterm:= newintprot ( intval ( pt ) )
	else begin
		p:= initprot ( mt [ pt ] );
		for k:= 1 to getarity ( pt ) do begin
			arg:= argument ( pt, env, k );
			if isndvar ( arg ) then
				newpvararg( p,k,varconvert( varadr( arg,env)))
			else newparg ( p, k, storeterm ( arg, env ) )
		end {for};
		storeterm:= wrapprot ( p )
	end
end;

function storeclause ( headpt, bodypt : mtx;  env : stackx )  : procptr;
{ headpt and bodypt are correct terms ( or prots in env ) representing
   a clause head and body ( cf.assert ).  store this clause.              }
var   p : procptr;
begin   vardictbot:= vtop;      { variable dictionary used by storeterm }
	new ( p );
	with p ^  do begin
		next:= nil;   systemroutine:= false;
		head:= storeterm ( headpt, env );
		body:= storeterm ( bodypt, env );
		nvars:= vtop - vardictbot       {no. of vars in dictionary}
	end;
	vtop:= vardictbot;   storeclause:= p
end;

function procok ( headatom : atx )  : boolean;
{ if headatom is associated with a system routine or protected procedure
   (but single-letter argless predicates can't be protected!),
   yield false & a warning                                                    }
var   ok, keep : boolean;
begin   with at [ headatom ] do begin
		if proc = nil then ok:= true
		else
		if proc ^. systemroutine then ok:= false
		else with proc^ do
			ok:= ( (head >= gprotect) and (head <= ngprotect) )
			     or ((head >= minchprotx) and (head <= maxchprotx));
		if not ok then begin
			keep:= telling;   telling:= false;   putline;
			write ( output, '-----redefinition of system routine ',
				    'or protected procedure ignored :'     );
			putline;   putname ( headatom );   putch( '/' );
			putint ( arity );   putline;   telling:= keep
		end
	end {with};
	procok:= ok
end;

procedure linkclause ( cl : procptr;  n : integer );
{ insert cl after the nth clause in its procedure ( as first if n <= 0 ,
   as last if n too large). ** note the effect of redefining ** }
var   headatom : atx;   p : procptr;
begin   headatom:= mt [ cl ^. head ];
	if procok ( headatom ) then begin
		if redefining and ( headatom <> lastasserted ) then begin
			lastasserted:= headatom;   at [ headatom ]. proc:= nil
		end;
		p:= at [ headatom ]. proc;
		if ( n <= 0 ) or ( p = nil ) then       {insert as first}
			with at [ headatom ] do begin
				cl^.next:= proc;   proc:= cl
			end
		else begin      { find clause n  (or last) to link to }
			while ( n <> 1 ) and ( p^.next <> nil ) do begin
				n:= n - 1;   p:= p^.next
			end;
			cl^.next:= p^.next;   p^.next:= cl
		end
	end
end {linkclause};

procedure assert;
{ implement  assert( head, body, number )  :
   construct a clause and invoke linkclause.    }
begin   if not is_normal(spar[1]) or not isint(spar[3]) then syserror:= true
	else
	if not iscallsequence ( spar [ 2 ], ancenv ) then syserror:= true
	else  linkclause ( storeclause( spar[1], spar[2], ancenv ), sparv[3] )
end;

procedure clause ( var success : boolean );
{ implement  clause( functor, arity, n, head, body) . the first 3 are
   given. find the n'th clause in the procedure associated with functor &
   arity (failing if none) and try to unify it with head and body.        }
var   atom : atx;   p : procptr;   k : integer;
begin   if not isint(spar[3]) then syserror:= true
	else atom:= atomspar ( 1, 2 );     { checks parameters, remember ? }
	if not syserror then begin                      { call ok }
		p:= at [ atom ]. proc;   k:= 1;
		while ( p<> nil ) and ( k < sparv [ 3 ] ) do begin
			p:= p^.next;   k:= k + 1
		end;
		if ( p = nil ) or ( sparv [ 3 ] < 1 ) then success:= false
		else
		if p^.systemroutine then success:= false
		else begin      { create fresh clause instance and unify }
			{check (11, topenv = vtop );}   pushvars ( p^.nvars );
			success:= unify( spar[ 4 ], ancenv, p^.head, topenv );
			if success then         {heads unified, bodies now }
				success:= unify ( spar [ 5 ], ancenv,
							p^.body, topenv )
		end
	end {call ok}
end {clause};

procedure genproc ( var success : boolean );
{ implement  $proc( head ) , an auxiliary of  proc( head )  ( which is
   written in prolog ) . try to unify with the most general term instance
   of the atom shown by procindx (cf. proclimit).                         }
begin   success:= unify ( spar [ 1 ], ancenv, gterm ( procindx ), vtop )
end;

procedure proclimit ( var success : boolean );
{ implement  $proclimit , a shameful auxiliary of  proc( head ) .
   set procindx (which is initialised by  $procinit , cf. sysroutcall)
   to the next atom associated with a non-system procedure and  f a i l .
   (succeed if no such atoms left).                                         }
var   found : boolean;
begin   found:= false;
	while ( procindx < athigh ) and not found do begin
		procindx:= procindx + 1;
		with at [ procindx ] do
			if name <> emptyatom then
				if proc <> nil then
					found:= not proc^.systemroutine
	end;
	success:= not found
end;

procedure retract ( var success : boolean );
{ implement  retract( name, arity, number ) :  retract the number'th
   clause of those whose head predicate has this name and arity,
   fail if none such.                                                   }
var   atom : atx;   p : procptr;   n : integer;
begin   atom:= atomspar ( 1, 2 );
	if not syserror then syserror:= not procok ( atom )  or
						not isint ( spar [ 3 ] );
	if not syserror then begin                              { call ok }
		n:= sparv [ 3 ];
		if n < 1 then success:= false
		else
		if n = 1 then           { remove first clause }
			with at [ atom ] do
				if proc = nil then success:= false
				else proc:= proc^.next
		else begin              { find the preceding clause }
			p:= at [ atom ]. proc;
			while ( n <> 2 ) and ( p <> nil ) do begin
				n:= n - 1;   p:= p^. next
			end;
			if p = nil then success:= false        {no (n-1)'th}
			else with p^  do
				if next <> nil then next:= next ^. next
		end {not 1st}
	end {call ok}
end {retract};

procedure abolish;
{ implement  abolish( name, arity )  }
var   atom : atx;
begin   atom:= atomspar ( 1, 2 );
	if not syserror then if procok ( atom ) then at [ atom ]. proc:= nil
end;

procedure redefine;
{ tag used as nil ( tag can't be redefined ) }
begin   redefining:= not redefining;
	if not redefining then lastasserted:= std [atmtag]
end;

procedure predefined ( var success : boolean );
{ implement  predefined( name, arity ) :  is this a system routine ? }
var   atom : atx;
begin   atom:= atomspar ( 1, 2 );
	if not syserror then
		with at[ atom ] do
			 if proc = nil then success:= false
			 else success:= proc ^. systemroutine
end;


procedure protect;
{ set protected boundaries, so that clauses which are already defined
   will be protected by procok                                          }
begin   gprotect:= gtop;   ngprotect:= ngbot
end;





{---------  h i g h  i n p u t / o u t p u t  ---------}



procedure rdchar;
{ read the next character into cch ( linend returned as chr(eol) ).
   non-user file end redirects input stream to user, user file end is fatal.}
begin   while endoffile do
		if sysloading or seeing then begin
			closefile ( true );   writeln ( output );
			writeln ( output, ' -----end of file, terminal input now' );
			putline
		end
		else errorr ( usereof );
	if endofline then begin
		getline;   cch:= chr ( eol )
	end
	else getch ( cch )
end {rdchar};

procedure lastch ( var success : boolean );
{ try to unify sysrout parameter with current character }
var   c : char;
begin   if isvar ( spar [ 1 ] ) then bindpar ( 1, charprot ( cch ) )
	else
	if ischar ( spar [ 1 ] , c ) then success :=  cch = c
	else success:= false
end;

procedure sskipbl;
{ ensure that cch is not filled with a white space character }
begin   while cch <= ' ' do rdchar              {** ASCII version **}
end;

procedure wch;
{ output the sysrout parameter, which should be a character term }
var   c : char;
begin   if ischar ( spar [ 1 ], c ) then
		if c = chr ( eol ) then putline
		else putch ( c )
	else syserror:= true
end;

procedure newfile ( see : boolean );
{ implement see or tell, according to parameter }
begin   if not isargless ( spar [ 1 ] ) then syserror:= true
	else
	if mt [ spar [ 1 ] ]  =  std [atmuser] then
		closefile ( see )
	else begin
		openfile ( at [ mt [ spar [ 1 ] ] ] . name ,  see );
		if see then seeobj:= spar [ 1 ]
		else tellobj:= spar [ 1 ]
	end
end {newfile};

procedure whatfile ( see : boolean;  var success : boolean );
{ implement seeing or telling, according to parameter }
var   fileterm : mtx;
begin   if see then fileterm:= seeobj
	else fileterm:= tellobj;
	success:= unify ( spar [ 1 ] , ancenv, fileterm, vtop )
end;




{---------  t e r m  i / o  -  o p s  ---------}

function getop ( nm : atx ) : opptr;
{ nm is a name :  return the associated op declaration or nil }
var   p : opptr;   found : boolean;
begin   found:= false;   p:= opht[nm mod ophashsize + 1];
	while ( p <> nil ) and not found do
		if p ^. name = nm then found:= true
		else p:= p ^. next;
	getop:= p                  { found or nil }
end;

procedure storeop ( op : opptr );
{ store a (new) op declaration record in its hash chain }
var   h : ophashx;
begin   h:= op ^. name mod ophashsize + 1;
	op ^. next :=  opht[h];   opht[h] := op
end;

function mkkind ( kinds : opkinds )  : opkind;
{ used when card(kinds)<=1. find the kind or noop if kinds empty. }
var   k : opkind;
begin   k:= xf;
	while not ( k in kinds ) and ( k <> noop ) do  k:= pred ( k );
	mkkind:= k
end;

function onekind ( kinds : opkinds )  : boolean;
{ an op has at most one binary and one unary kind. true if exactly one kind. }
begin   onekind:= ( kinds <> [] ) and
		  ( ( kinds*unaryops = [] ) or ( kinds*binaryops = [] ) )
end;

function isanop ( nm : atx;  arity : argnumb;
		  var kind : opkind;  var prior : oppriority ) : boolean;
{ return op declaration info for this name / arity, false if not an op at all}
var   op : opptr;   kinds : opkinds;
begin   op:= getop ( nm );
	if op = nil then begin
		kind:= noop;   prior:= nopriority;   isanop:= false
	end
	else begin
		prior:= op ^. priority;
		if arity = 2 then kinds:= op^.kinds * binaryops
		else
		if arity = 1 then kinds:= op^.kinds * unaryops
		else kinds:= [];
		{ at this point card(kinds)<=1 due to op decl constraints }
		kind:= mkkind ( kinds );   isanop:= true
	end
end;

procedure wrop ( nm : atx;  kinds : opkinds;  prior : oppriority );
{ output op information (used for various messages) }
var   k : opkind;
begin   putname ( nm );   putch ( ' ' );                          { the name }
	for k:= fx to xf do                                 { the patterns }
		if k in kinds then begin
			putname ( opatoms[k] );   putch ( ' ' )
		end;
	putint ( prior )                                    { the priority }
end;

procedure opredefmess ( nm : atx;  kinds : opkinds;  prior : oppriority );
{ output a warning about op redefinition }
var   keep : boolean;
begin   keep:= telling;   telling:= false;
	putline;   write ( output, ' --- op redefined : ' );
	wrop ( nm, kinds, prior );   putline;
	telling:= keep
end;

procedure newop;
{ implement  op( priority, kind, name )
   conflicts with existing declarations  are resolved through redefinition
   (of the whole op if priority conflict, of unary/binary class if only
   new kind defined). duplicated declarations have no effect.}
var   nm : atx;   op : opptr;   kind : opkind;   kinds : opkinds;
begin   if not isint ( spar[1] ) or not isargless ( spar[2] ) or
			      not isargless ( spar[3] ) then syserror:= true
	else
	if ( sparv[1] < minpriority ) or ( sparv[1] > maxoppriority ) then
		syserror:= true
	else begin                         { superficially, the call is ok }
		nm:= mt[ spar[3] ];                         { op's name }
		kind:= xf;           { search for pattern matching 2nd arg }
		while (kind<>noop) and (opatoms[kind]<>mt[spar[2]]) do
			kind:= pred ( kind );
		{if pattern unknown or name needs quoting or
		  name is ',' or '|' then not accepted       }
		if ( kind = noop ) or needsquoting ( nm ) or
		   ( nm = commaopname ) or ( nm = baropname ) then
			 syserror:= true
		else begin
			if kind in unaryops then kinds:= unaryops
			else kinds:= binaryops;                      { class }
			op:= getop ( nm );         { already an op ? }
			if op = nil then begin     {no! create, fill in later}
				new ( op );   op^.name:= nm;
				op^.kinds:= [];   storeop ( op )
			end
			else                      { yes ! check for conflict }
			if op^.priority <> sparv[1] then begin
				opredefmess ( nm, op^.kinds, op^.priority );
				op^.kinds:= []     { complete redefinition }
			end
			else
			if ( kinds * op^.kinds <> [] ) and
					 not ( kind in op^.kinds ) then begin
				opredefmess ( nm, kinds * op^.kinds,
								op^.priority );
				op^.kinds:= op^.kinds - kinds  {class redef.}
			end;
			op^.priority:= sparv[1];
			op^.kinds:= op^.kinds + [kind]
		end {call ok}
	end {superficially ok}
end {newop};

procedure delop;
{ implement  delop( name ) :  delete op declaration for this name, if any
  (by resetting its name to that of '(').                                      }
var   op : opptr;
begin   if not isargless ( spar[1] ) then syserror:= true
	else begin
		op:= getop ( mt[ spar[ 1 ] ] );
		if op <> nil then op^.name:= mt[ charprot( '(' ) ]
	end
end {delop};





{---------  t e r m   i / o  -  w r i t e r  ---------}

	{ throughout this section, q is true if output
	   is to be readable for Prolog (cf. writeq)    }
	{ writer uses temporary variable dictionary above
	   variable stack ( cf. clause/assert/retract )    }


procedure wr ( t : mtx;  env : stackx;  q : boolean;
	       leftarg : boolean;  inkind : opkind;  inprior : oppriority );
forward;  {- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }


procedure wrname ( nm :atx;   q : boolean );
{ output this name }
begin   if not q then putname ( nm )
	else
	if needsquoting ( nm ) then putqname ( nm )
	else putname ( nm )
end;

procedure wrstd ( t : mtx;  env : stackx;  isanop : boolean;  q : boolean );
{ write a (normal) term in standard format, isanop forces quoting }
var   arity, k : argnumb;   name : atx;
      arg : mtx;
begin   name:= mt[t];   arity:= at[name].arity;
	if isanop then putqname ( name )
	else wrname ( name, q );
	if arity <> 0 then begin
		putch ( '(' );   putch ( ' ' );
		for k:= 1 to arity do begin
			arg:= argument ( t, env, k );
			{ ops at and above commapriority must be bracketed
			   to avoid incorrect reparsing when read in }
			wr ( arg, env, q, false, xfx, commapriority );
			if k <> arity then putch ( ',' );
			putch ( ' ' )
		end;
		putch ( ')' )
	end {arity<>0}
end {wrstd};

function isstring ( t : mtx;  env : stackx ) : boolean;
{ is this term a (closed) list of characters ?  its main functor is a dot ! }
var   noncharseen : boolean;   next : mtx;   dummy : char;
begin   noncharseen:= false;
	repeat  if not ischar ( argument ( t, env, car ), dummy ) then
			noncharseen:= true
		else begin  {  alas, Pascal has no conditional and }
			next:= argument ( t, env, cdr );
			if is_normal ( next ) then t:= next
			else noncharseen:= true              { a var or int }
		end
	until ( mt[t] <> std[atmdot] ) or noncharseen;
	isstring:=  mt[t] = std[atmnil]
end;

procedure wrstring ( t : mtx;  env : stackx );
{ the term is a string ( cf. isstring ) :  output it }
var   c : char;
begin   putch ( '"' );
	while mt[t] = std[atmdot] do begin
		c:= ct[at[mt[argument(t,env,car)]].name];
		putch ( c );
		if ccclass[ c ]= cstringquote then putch ( c );   { "" }
		t:= argument ( t, env, cdr )
	end;
	putch ( '"' )
end;

procedure wrlist ( t : mtx;  env : stackx;  q : boolean );
{ the term is a dot :  output a list (possibly a string) }
var   stop,   nilseen : boolean;   arg : mtx;   nested : boolean;
begin   if isstring ( t, env ) then wrstring ( t, env )
	else begin
		putch ( '[' );   nilseen:= false;   stop:= false;
		repeat  arg:= argument ( t, env, car );   nested:= false;
			if is_normal ( arg ) then
				nested:= mt[ arg ] = std[ atmcomma ];
			if nested then putch ( '(' );
			wr( arg, env, q, false, noop, nopriority );
			if nested then putch ( ')' );
			t:= argument ( t, env, cdr );
			if not is_normal ( t ) then stop:= true
			else
			if mt[ t ] <> std[ atmdot ] then begin
				stop:= true;
				nilseen:= mt[ t ] = std[ atmnil ]
			end;
			if not stop then begin
				putch ( ',' );   putch ( ' ' )
			end
		until stop;
		if not nilseen then begin               { a bar is needed }
			putch ( ' ' );   putch ( '|' );   putch ( ' ' );
			wr ( t, env, q, false, noop, nopriority )
		end;
		putch ( ']' )
	end
end {wrlist};

procedure wr {( t : mtx;  env : stackx;  q : boolean;
		 leftarg : boolean;  inkind : opkind;  inprior : oppriority )};
{ temporary variable dictionary used !
   write the term according to current op declarations. leftarg is true if
   this is the left argument of a postfix or infix functor. inkind and
   inprior carry information about the immediately surrounding term.        }
var   nm : atx;   ar : argnumb;   kind : opkind;   prior : oppriority;
      nested : boolean;
begin   if isint ( t ) then putint ( intval ( t ) )
	else
	if isndvar ( t ) then begin
		putch ( 'X' );   putint ( varconvert ( varadr ( t, env ) ) )
	end
	else
	if t = dumvarx then putch ( '_' )
	else                                                { a normal term }
	if mt[t] = std[atmdot] then wrlist ( t, env, q )
	else begin                                     { normal, not a list }
		ar:= at[ mt[ t ] ].arity;
		nm:= samename( mt[ t ], 0 );
		if not isanop( nm,ar,kind,prior ) then wrstd( t,env,false,q )
		else          { kind & prior were defined by the call above }
		if kind = noop then wrstd ( t , env, q, q )  {op used as noop}
		else begin                     { an op! parentheses needed? }
			if inkind = noop then nested:= false
			else
			if prior < inprior then nested:= false
			else
			if prior > inprior then nested:= true
			else                                { same priority }
			if leftarg then nested:= ( kind in rightops ) or
						       not ( inkind in leftops )
			else nested:= ( kind in leftops ) or
						     not ( inkind in rightops );
			if nested then putch ( '(' );
			{ note that ops are readable sans quotes }
			if kind in prefixops then begin
				putname( nm );   putch ( ' ' )
			end;
			wr ( argument( t, env, 1 ), env , q,
				   not ( kind in prefixops ), kind, prior );
			if not ( kind in prefixops ) then begin
				putch ( ' ' );   putname ( nm )
			end;
			if kind in infixops then begin
				putch ( ' ' );
				wr( argument(t,env,2),env,q,false,kind,prior )
			end;
			if nested then putch ( ')' )
		end{an op}
	end{not a list}
end {wr};

procedure dowrite { ( term : mtx;   q : boolean ) };
{ implement  write( term )  or  writeq( term )  , according to q.
   use temporary variable dictionary (cf. clause/assert/retract).
   parameter not passed through spar, because dowrite is also called
   directly to implement debug.                                       }
begin   vardictbot:= vtop;
	wr ( term, ancenv, q, false, noop, nopriority );
	vtop:= vardictbot
end;





{---------   #######  t e r m   r e a d e r ####### ---------}
{            #######                        #######
		scanner and reader are immersed in
		procedure doread to allow efficient
		error recovery (a non-local goto)
		without exiting invocation context.
	      #######                        #######           }

procedure doread ( fromsyscall : boolean;  var success : boolean );
{ fromsyscall  is false if called by the main loop reader -
   readtogoal is executed then to fill goalstmnt & ngoalvars. }
label 3;
var   term : mtx;

procedure synterr ( id : synterrid );
forward; {------------------------}


{---------  t e r m   i / o  -  s c a n n e r  ---------}

	 { scanner uses temporary variable dictionary above variable stack
	    (but handled differently than in clause/assert/retract) and
	    a temporary string area in high character table.               }


procedure wrlex ( lex : token );
{ output this token. used for various messages. }
begin   with lex do
		case class of
			noarglex   : display ( term, vtop );
			idarglex   : begin   putname ( idname );
					     putch ( '/' );  putint ( idarity )
				     end;
			oplex      : wrop ( opname, kinds, prior );
			lroundlex  : putch ( '(' );
			lcurlylex  : putch ( '{' );
			lsquarelex : putch ( '[' );
			barlex     : putch ( '|' );
			rsquarelex : putch ( ']' );
			rcurlylex  : putch ( '}' );
			rroundlex  : putch ( ')' );
			dotlex     : begin   putch( ' ' );   putch( '.' )  end
		end{case};
	putch ( ' ' )
end {wrlex};

procedure rd;
{ read next character, skipping end of line. fill cc with cch's class. }
begin   repeat  rdchar
	until cch <> chr( eol );
	cc:= ccclass[ cch ]
end;

procedure skipcomment;
{ skip chars upto end of line and get first char of new line }
begin   while cch <> chr( eol ) do rdchar;
	rd
end;

procedure skipblanks;
{ cc is a blank, read in something different }
begin   repeat  rd
	until cc <> cblank
end;

procedure processid ( quoted : boolean );
{ an id was just built :  wrap it and build noarg, op or idarg token.
   not op if quoted !                                                    }
var   op : opptr;   nm : atx;
begin   nm:= findatom ( wrapname, 0 );
	if quoted then op:= nil
	else op:= getop ( nm );
	with inlex do
		if op <> nil then begin
			class:= oplex;   opname:= nm;
			prior:= op^.priority;   kinds:= op^.kinds
		end
		else                                             { not an op }
		if cc = clbracket then begin                       { id( }
			class:= idarglex;   idname:= nm;   idarity:= 1;
			rd
		end
		else begin
			class:= noarglex;   term:= gterm ( nm )
		end
end {processid};

procedure rdid;
{ the symbol is an identifier, cch at first letter }
begin   repeat  buildname ( cch );   rd
	until cc > cunderscore;
	processid ( false )
end;

procedure rdqid;
{ the symbol is a quoted id, cch at idquote. two idquotes count as one. }
begin   repeat  rd;
		while cc <> cidquote do begin
			buildname ( cch );   rd
		end;
		rd;
		if cc = cidquote then buildname ( cch )
	until cc <> cidquote;
	processid ( true )
end {rdqid};

procedure rdsolo;
{ a single-char id symbol }
begin   buildname ( cch );   rd;
	processid ( false )
end;

procedure rdsym;
{ an id symbol. first char possibly processed already.
   note that a period immediately following a symch is treated as symch. }
begin   while cc in [csymch, cperiod, csign] do begin
		buildname ( cch );   rd
	end;
	processid ( false )
end;

procedure rdstring;
{ the symbol is a string, cch at stringquote. two quotes are one inside.
   characters are first pushed into ct, then unstacked to form a list.
   NOTE the direct manipulation of ctop and knowledge of buildname.          }
var   l : mtx;   save : ctx;
begin   save:= ctop;
	repeat  rd;
		while cc <> cstringquote do begin
			buildname ( cch );   rd
		end;
		rd;
		if cc = cstringquote then buildname ( cch )
	until cc <> cstringquote;
	l:= nilprotx;
	while ctop <> save do begin
		ctop:= ctop - 1;   l:= heapdot ( charprot( ct[ctop] ), l )
	end;
	inlex.class:= noarglex;   inlex.term:= l
end {rdstring};

procedure rdint ( sign : integer );
{ sign is 1 or -1, cch is a digit. read the integer.
   overflow checking as in trydigseq :
   in two's complement smallest integer will be out of range.}
var   v, d : integer;
begin   v:= 0;
	repeat  d:= ord( cch ) - ord( '0' );
		if v > maxint div 10 then synterr ( bigint );
		v:= v * 10;
		if v <> 0 then
			if ( maxint div v = 1 ) and ( maxint mod v < d ) then
				synterr ( bigint );
		v:= v + d;   rd
	until cc <> cdigit;
	inlex.class:= noarglex;   inlex.term:= heapint ( sign * v )
end {rdint};


{ the following two procedures convert variable names into variables.
   variable names are stored in high ct (above cbot) and are located
   by linear search through pairs stored in low stack area above
   vardictbot. the first element of each pair is a pointer to the lowest
   char of a variable name (which is stored reversed), the second shows
   a copy stack variable associated with this name.                      }

procedure buildvar ( c : char );
begin   cspace;   ct[ cbot ]:= c;   cbot:= cbot - 1
end;

function wrapvar : stackx;
var   v : stackx;   found : boolean;
begin   v:= vardictbot;   found:= false;
	while ( v < vtop ) and not found do
		if compstrings ( cbot + 1, mt[ v ], true ) then found:= true
		else v:= v + 2;
	if v <> vtop then cbot:= builtvar    {found: discard repeated string}
	else begin
		stackspace ( 3 );            { 2 for entry, 1 for var }
		mt[ v ]:= cbot + 1;   mt[ v + 1 ]:= csbot;
		mt[ csbot ]:= freevar;   csbot:= csbot - 1;   vtop:= vtop + 2;
		{prepare for next var name}
		buildvar ( chr( eos ) );   builtvar:= cbot
	end;
	wrapvar:= mt[ v + 1 ]
end;

procedure rdvar;
{ cch is a bigletter. read the variable. }
begin   repeat  buildvar ( cch );   rd
	until cc > cunderscore;
	inlex.class:= noarglex;   inlex.term:= wrapvar
end;

procedure rdlbracket;
{ cch is a left bracket. read it. }
begin   with inlex do begin
		if cch = '(' then class:= lroundlex
		else
		if cch = '[' then begin
			class:= lsquarelex;   listlength:= 1
		end
		else
		if cch = '{' then class:= lcurlylex	{ CURLY }
	end;
	rd
end;

procedure rdrbracket;
{ cch is a right bracket. read it. }
begin   with inlex do begin
		prior:= maxpriority;
		if cch = ')' then class:= rroundlex
		else
		if cch = ']' then class:= rsquarelex
		else
		if cch = '}' then class:= rcurlylex	{ CURLY }
	end;
	rd
end;

procedure rdsign;
{ cch is '+' or '-'. read the symbol or integer. }
var   signch : char;   signval : -1..1;
begin   if cch = '-' then signval:= -1
	else signval:= 1;
	signch:= cch;   rd;
	if cc = cdigit then rdint ( signval )
	else begin           { do the first char for rdsym }
		buildname ( signch );   rdsym
	end
end {rdsign};

procedure rdperiod;
{ cch is a period. token is sym  or dot. }
begin   rdchar;   cc:= ccclass[ cch ];         { don't want eol skipped }
	if cc = cblank then begin
		inlex.class:= dotlex;   inlex.prior:= maxpriority
	end
	else begin          { process the first char for rdsym }
		buildname ( '.' );   rdsym
	end
end {rdperiod};

procedure scan;
{ uses temporary areas in high ct & low (var) stack !
   fill inlex with the next input symbol.
   cch contains unprocessed character, cc is its class. }
var   oncemore : boolean;          { forces repetition for comments & blanks }
begin   inlex.prior:= nopriority;     { most likely outcome }
	repeat  oncemore:= false;
		case cc of
			cbigletter   : rdvar;
			csmalletter  : rdid;
			cdigit       : rdint ( 1 );
			cunderscore  : with inlex do begin
					    class:= noarglex;
					    term:= dumvarx;   rd
				       end;
			clbracket    : rdlbracket;
			crbracket    : rdrbracket;
			csolochar    : rdsolo;
			csymch       : rdsym;
			cidquote     : rdqid;
			cstringquote : rdstring;
 			cpercent     : begin     { a comment }
					   oncemore:= true; skipcomment
				       end;
			cperiod      : rdperiod;
			csign        : rdsign;
			cbar         : with inlex do begin
					   class:= barlex;
					   prior:= maxpriority;   rd
				       end;
			cblank       : begin  oncemore:= true; skipblanks
				       end;
			cother       : synterr ( badchar )
		end{case}
	until not oncemore
end {scan};

procedure startscanning;
begin   buildvar ( chr( eos ) );   builtvar:= cbot;
	vardictbot:= vtop;
	cc:= ccclass[ cch ];   scan
end;

procedure stopscanning;
begin   cbot:= cthigh;   vtop:= vardictbot
end;



{---------  t e r m   i / o   -   p a r s e r  ---------}

	{ parser uses the free area above ground prototypes
	   for its argument stack.
	   cf. scanner for other dirty overlays.             }



procedure pushparsel;
{ save toplex in  the cellar }
var   p : parselptr;
begin   if freeparsels = nil then begin
		new ( freeparsels );   freeparsels^.prev:= nil
	end;
	p:= freeparsels;   freeparsels:= p^.prev;
	p^.lex:= toplex;
	p^.prev:= cellar;   cellar:= p
end;

procedure popparsel;
{ restore toplex from the cellar }
var   p : parselptr;
begin   p:= cellar;   cellar:= p^.prev;
	toplex:= p^.lex;
	p^.prev:= freeparsels;   freeparsels:= p
end;

procedure pusharg ( term : mtx );
{ push a term onto the argument stack }
begin   protspace ( 1 );
	mt[ gtop ]:= term;   gtop:= gtop + 1
end;

procedure closearguments ( name : atx;  arity : argnumb );
{ build a term for a functor of this name and arity,
   using arguments from arg stack & replacing them with a pointer to the term }
var   t : mtx;   k : argnumb;
begin   t:= gterm ( samename ( name, arity ) );
	for k:= arity downto 1 do begin     { pointers to args, cf. terminst }
		gtop:= gtop - 1;   mt[ t + k ] :=  - mt[ gtop ]
	end;
	pusharg ( t )
end;

procedure shift;
{ note that shiftactions is normally used to treat ops }
begin   pushparsel;   toplex:= inlex;
	justreduced:= false;   scan
end;

procedure reduce;
{ called only for noargs & ops (which are certainly corect & nonambiguous) }
begin   with toplex do
		if class = noarglex then pusharg ( term )
		else
		if class <> oplex then synterr ( bug )
		else
		if kinds * binaryops = [] then closearguments( opname, 1 )
		else closearguments ( opname, 2 );
	popparsel;   justreduced:= true
end;

procedure shiftactions;
{ shift is called for, but is it correct ? special treatment of comma
   in lists & argument lists, partial op disambiguation, change of
   priority for idarg & left brackets.                                 }
var   comma : boolean;
begin   comma:= false;
	with inlex do begin
		if justreduced and (class<>oplex) and (class<>barlex) then
			synterr ( nodelim );
		if class = oplex then begin      { disambiguate if possible }
			if justreduced then kinds:= kinds - prefixops
			else kinds:= kinds * prefixops;
			if kinds = [] then synterr ( opbad );
			{ comma treated }
			if opname = commaopname then
			      with toplex do
				  if class in [idarglex,lsquarelex] then begin
				      comma:= true;
				      if class = idarglex then
					  idarity:= idarity + 1
				      else listlength:= listlength + 1
				  end
		end{oplex}
		else
		if class in [idarglex,lroundlex,lcurlylex,lsquarelex] then
			prior:= maxpriority
	end{with};
	if not comma then shift
	else begin
		justreduced:= false;   scan
	end
end {shiftactions};

procedure redactions;
{ reduce is called for, but is it correct ?
   disambiguate ops on stack. called only for noargs & ops. }
begin   with toplex do
		if class = oplex then begin
			if justreduced then kinds:= kinds - postfixops
			else kinds:= kinds * postfixops;
			if not onekind ( kinds ) then synterr ( opbad )
		end;
	reduce
end {redactions};

procedure closeidarg;
{ inlex is a ')' matching 'id('. toplex.idarity args reduced. }
begin   if not justreduced then synterr ( noarg );
	closearguments ( toplex.idname, toplex.idarity )
end;

procedure closelist ( tail : mtx );
{ '[' - ']' pair seen, list elements on arg stack, toplex.listlength
   indicates how many. form a list of these and tail, replace elements
   on arg stack with a pointer to the list                             }
var   k : integer;
begin   if savegtop + toplex.listlength > gtop then synterr ( bug );
	for k:= toplex.listlength downto 1 do begin
		gtop:= gtop - 1;   tail:= heapdot ( mt[gtop], tail )
	end;
	pusharg ( tail )
end {closelist};

procedure opeqop;
{ two ops of equal priorities seen.
   if ambiguous, raise synterr; otherwise shift or reduce. }
const   un = false;   bin = true;                    { indexes in x }
var   x : array[ boolean, boolean ] of boolean;    { true is a conflict }
      ktop, kin : opkind;   action :opopaction;
begin   { gather conflict info for all kind combinations }
	x[ un, un ] :=  true;   x[ un, bin ] :=  true;
	x[ bin, un ] :=  true;   x[ bin, bin ] :=  true;
	for ktop:= fx to xf do
	  if ktop in toplex.kinds then
	    for kin:= fx to xf do
	      if kin in inlex.kinds then
		x[ ktop in binaryops, kin in binaryops ] :=
		not((justreduced and(opop[ktop,kin]in[shiftifarg,redifarg]))
		     or(not justreduced and
			      (opop[ktop,kin]in[shiftifnoarg,redifnoarg])));
	{ remove kinds conflicting in all combinations }
	with toplex do begin
		if x[un,un] and x[un,bin] then kinds:= kinds - unaryops;
		if x[bin,un] and x[bin,bin] then kinds:= kinds - binaryops
	end;
	with inlex do begin
		if x[un,un] and x[bin,un] then kinds:= kinds - unaryops;
		if x[un,bin] and x[bin,bin] then kinds := kinds - binaryops
	end;
	if (inlex.kinds=[]) or not onekind(toplex.kinds) then synterr(opbad);
	{ two actions possible now, if input still ambiguous }
	ktop:= mkkind ( toplex.kinds );
	if onekind(inlex.kinds) then action:= opop[ktop, mkkind(inlex.kinds)]
	else begin                        { see if both actions are the same }
		action:= opop[ ktop, mkkind( inlex.kinds * binaryops ) ];
		if action <> opop[ktop, mkkind( inlex.kinds * unaryops )] then
			synterr ( ambig )
	end;
	if action in [ redifarg, redifnoarg ] then reduce
	else shift
end {opeqop};

procedure eqpriority;
{ both priorities equal }
var action : eqpraction;
begin   if inlex.prior = nopriority then synterr ( nodelim )      { 2 noargs }
	else
	if inlex.prior <> maxpriority then opeqop                 { two ops }
	else begin
		action:= eqpr[ toplex.class, inlex.class ];
		case action of
		    badmatch       : synterr ( nomatch );
		    shiftbar       : if not justreduced then synterr( badbar )
				     else shift;
		    closeidcumargs : closeidarg;
		    closenilorlist : if justreduced then closelist( nilprotx )
				     else pusharg( nilprotx );       { [] }
		    closebarlist   : if not justreduced then synterr( badbar )
				     else begin
					 popparsel;         {uncovering '['}
					 gtop:= gtop - 1;   {bar's 2nd arg}
					 closelist(mt[gtop])    {a trick}
				     end;
		    closeround     : if not justreduced then
					 synterr( roundempty );
		    closecurly     : if not justreduced then
					 synterr( curlyempty )
				     else closearguments( std[atmcurly], 1 )
		end{case};
		if action <> shiftbar then begin     { remove the brackets }
			justreduced:= true;   popparsel;   scan
		end
	end{maxpriority}
end {eqpriority};

procedure parse ( var term : mtx );
{ a customized operator precedence parser. op expressions considered
   ambiguous if two neighbouring ops of equal priority (possibly
   separated by an argument) have two interpretations consistent
   with their declarations (only mixed ops may cause trouble).        }
begin   savegtop:= gtop;  startscanning;
	justreduced:= false;
	while inlex.class <> dotlex do
		if toplex.prior > inlex.prior then shiftactions
		else
		if toplex.prior < inlex.prior then redactions
		else eqpriority;
	{ unstack remaining elements }
	while toplex.class <> dotlex do
		if toplex.class in [oplex, noarglex] then redactions
		else synterr ( unclosed );
	if savegtop = gtop then synterr ( empty );
	if savegtop <> gtop - 1 then synterr ( bug );
	term:= mt[ savegtop ];
	stopscanning;   gtop:= savegtop
end {parse};


{--------- #########  doread ######### ---------}

procedure synterr { ( id : synterrid ) };
var   keeptelling, keepecho : boolean;
begin   keeptelling:= telling;   telling:= false;
	putline;   write ( output, '++++++++ read error :' );   putline;
	case id of
	  bug         : write(output,'parser in error, show to maintenance crew');
	  badchar     : write(output,'bad character: ''', cch, ''' (ord=',
								 ord(cch), ')');
	  bigint      : write(output,'integer out of range');
	  opbad       : write(output,'op in ambiguous or incorrect context');
	  ambig       : write(output,'ambiguous mixed op');
	  nodelim     : write(output,'op or comma expected');
	  nomatch     : write(output,'brackets don''t match');
	  noarg       : write(output,'empty argument list');
	  badbar      : write(output,'bar in bad context');
	  roundempty  : write(output,'empty ''( )''');
	  curlyempty  : write(output,'empty ''{ }''');
	  unclosed    : write(output,'construct not terminated');
	  empty       : write(output,'empty term (dot only)')
	end{case};
	putline;   write(output,'context - stack/input/skipped :');   putline;
	wrlex ( toplex );   putline;   wrlex ( inlex );   putline;
	keepecho:= echo;   echo:= true;
	if id = badchar then rd
	else write( output, cch );
	while inlex.class <> dotlex do scan;
	putline;   echo:= keepecho;   telling:= keeptelling;
	{ wind up }
	stopscanning;   gtop:= savegtop;
	while toplex.class <> dotlex do popparsel;
	syserror:= true;   goto 3
end {synterr};

procedure readtogoal ( var goal : mtx;  var nvars : varnumb );
{ this is used by the interpreter to read the bootstrap monitor.
   each term read must be a list, which is treated as a clause.
   if the head is atmgo ( i.e. # ) the clause is executed as a goal. }
var   list : mtx;   clause : procptr;   isgoal : boolean;
begin   isgoal:= false;
	repeat  parse ( list );   vtop:= mtlow;  csbot:= stackhigh;
		if not iscallsequence ( list, vtop ) then errorr ( syntax );
		if mt[ list ] <> std[ atmdot ] then errorr ( syntax );
		clause:= storeclause ( argument ( list, vtop, car),
				       argument ( list, vtop, cdr ), vtop );
		if mt[ clause^.head ] = std[ atmgo ] then isgoal:= true
		else linkclause ( clause, maxint )
	until isgoal;
	goal:= clause^.body;   nvars:= clause^.nvars
end;


begin {doread}
	if not fromsyscall then readtogoal ( goalstmnt, ngoalvars )
	else begin
		parse ( term );
		success:= unify ( spar[1], ancenv, term, vtop )
	end;
3: end {doread};






{---------  s y s t e m  r o u t i n e s  ---------}



procedure haltaction ( var stop : boolean );
{ display the sysparam and halt the interpreter }
begin   putline;   display ( spar [ 1 ] , ancenv );   putline;
	stop:= true;   terminate:= true
end;

procedure sysroutcall { ( id : sysroutid;  var success, stop : boolean ) };
{ perform a system routine call }
var   k : nsysparam;
begin   syserror:= false;   success:= true;     { might change yet }
	for k:= 1 to getarity ( ccall ) do begin
		spar [ k ] :=  argument ( ccall, ancenv, k );
		if isint( spar [ k ] ) then sparv [ k ]:= intval( spar [ k ] )
	end;
	case id of
	idfail          : success:= false;      { keep this as first  }
	idtag ,
	idcall          : ;             { never called ! (cf. control) }
	idslash         : slash;
	idtagcut        : tagcut ( success );
	idtagfail       : tagfail ( success );
	idtagexit       : tagexit ( success );
	idancestor      : ancestor ( success );
	idhalt          : haltaction ( stop );
	idop            : newop;
	iddelop         : delop;
	idwrite         : dowrite ( spar[ 1 ], false );
	idwriteq        : dowrite ( spar[ 1 ], true );
	idread          : doread ( true, success );
	idstatus        : status;
	iddisplay       : display ( spar [ 1 ], ancenv );
	idrch           : rdchar;
	idlastch        : lastch ( success );
	idskipbl        : sskipbl;
	idwch           : wch;
	idecho          : echo:= true;
	idnoecho        : echo:= false;
	idsee           : newfile ( true );
	idseeing        : whatfile ( true, success );
	idseen          : closefile ( true );
	idtell          : newfile ( false );
	idtelling       : whatfile ( false, success );
	idtold          : closefile ( false );
	idordchr        : ordchr ( success );
	idsum           : sum ( success );
	idprod          : prod ( success );
	idinf           : inf ( success );
	idnameinf       : nameinf ( success );
	idbigletter     : ch1class ( cbigletter, success );
	idsmalletter    : ch1class ( csmalletter, success );
	idletter        : ch2class ( [csmalletter,cbigletter], success );
	iddigit         : ch1class ( cdigit, success );
	idalphanum      : ch2class (
			      [cbigletter,csmalletter,cdigit,cunderscore],
								      success );
	idbracket       : ch2class ( [clbracket,crbracket], success );
	idsolochar      : ch1class ( csolochar, success );
	idsymch         : ch2class ( [csymch,cperiod,csign], success );
	ideqvar         : eqvar ( success );
	idvar           : success:= isvar ( spar [ 1 ] );
	idatom          : success:= isargless( spar [1] );
	idinteger       : success:= isint ( spar [ 1 ] );
	idnonvarint     : success:= is_normal ( spar[1] );
	idfunctor       : functor ( success );
	idarg           : arg ( success );
	idpname         : pname ( success );
	idpnamei        : pnamei ( success );
	idproc          : genproc ( success );
	idproclimit     : proclimit ( success );
	idprocinit      : procindx:= atlow - 1;
	idclause        : clause ( success );
	idretract       : retract ( success );
	idabolish       : abolish;
	idassert        : assert;
	idredefine      : redefine;
	idpredefined    : predefined ( success );
	idprotect       : protect;
	idnonexistent   : shownonexistent:= true;
	idnononexistent : shownonexistent:= false;
	iddebug         : debug:= true;
	idnodebug       : debug:= false;
	idtrue          : {no action}         { keep this as last  }
	end {case}
end {sysroutcall};





{---------  i n i t i a l i s a t i o n  ---------}



function rdatom : atx;
{ used for kernel initialisation. format : 'name'/0 }
var   v : integer;
begin   while cch <> '''' do rdchar;
	rdchar;
	while cch <> '''' do begin
		buildname ( cch );   rdchar
	end;
	rdchar;
	if cch <> '/' then errorr ( loadfile );
	rdchar;
	if ccclass[ cch ] <> cdigit then errorr ( loadfile );
	v:= 0;
	repeat  v:= v * 10 + ( ord( cch ) - ord( '0' ) );
		rdchar
	until ccclass[ cch ] <> cdigit;
	rdatom:= findatom ( wrapname, v )
end;

procedure initvars;
{ initialise global variables (cf. startresolution & loadsyskernel) }
var   a : atx;  h : ophashx;
begin
	outlinesize:= 0;   solinesize:= 0;
	seeing:= false;   telling:= false;
	ctop:= ctlow;   cbot:= cthigh;   builtname:= ctop;
	freeatoms:= athigh - atlow + 1;
	for a:= atlow to athigh do at [ a ]. name:= emptyatom;
	ct [ emptyatom ] :=  chr ( eos );
	stackhigh:= mtlow - 1 + stacksize;   protlow:= stackhigh + 1;
	dumvarx:= protlow;   nilprotx:= dumvarx + 1;
	minchprotx:= nilprotx + 1;   maxchprotx:= minchprotx + maxchar;
	stdintprotx:= maxchprotx + 1;
	if stdintprotx+(maxstdint-minstdint) >= mthigh then errorr ( sysinit );
	mt [ dumvarx ] :=  varlim;
	gtop:= stdintprotx;     { loadsyskernel picks gtop up from here }
	ngbot:= mthigh;
	gprotect:= gtop;   ngprotect:= ngbot;
	tminfree:= tthigh;   bminfree:= bthigh;
	fminfree:= fthigh;   stackminfree:= stackhigh - mtlow;
	cminfree:= cthigh - ctlow;   pminfree:= mthigh - protlow;
	vtop:= mtlow;   csbot:= stackhigh;           { the rest in initres }
	terminate:= false;
	redefining:= false;
	shownonexistent:= false;      { don't warn on calls to nonexistent }
	cch:= ' ';   echo:= false;   debug:= false;
	{ character classification, ASCII version.
	   symch is a non-alphameric freely used to form symbols.  }
	for h:= 1 to ophashsize do opht[h]:= nil;
	binaryops   := [ xfy, xfx, yfx ];
	unaryops    := [ fx, fy, yf, xf ];
	rightops    := [ fy, xfy ];
	leftops     := [ yfx, yf ];
	prefixops   := [ fx, fy ];
	infixops    := binaryops;
	postfixops  := [ yf, xf ];
	with toplex do begin     { initialise parse stack bottom }
		prior:= maxpriority;   class:= dotlex
	end;
	cellar:= nil;   freeparsels:= nil;
	inlex:= toplex;        { initialise current token }
end {initvars};

procedure loadsyskernel;
{ create standard objects.
   kernel library will be processed in main program loop.    }
var   k : integer;   a : stdatomid;   s : sysroutid;   c : char;
      p : procptr;   prot : protx;   op, op2 : opkind;   pop : opptr;
      lex1, lex2 : lexclass;
begin   sysloading:= true;
	reset ( mc, 'sysfile' );
	readln ( mc );                                 { load ccclass table }
	for c:= chr( 0 ) to chr ( 63 ) do read ( mc, ccclass[ c ] );
	readln ( mc );
	for c:= chr ( 64 ) to chr ( 127 ) do read ( mc, ccclass[ c ] );
	readln ( mc );
(********** character codes should give the following effect :
	for c:= chr ( 0 ) to ' ' do ccclass[ c ]:= cblank;
	for c:= succ ( ' ' ) to chr ( maxchar ) do ccclass [ c ] := cother;
	for c:= 'a' to 'z' do ccclass [ c ] := csmalletter;
	for c:= 'A' to 'Z' do ccclass [ c ] := cbigletter;
	for c:= '0' to '9' do ccclass [ c ] := cdigit;
	ccclass [ '(' ] := clbracket;   ccclass [ ')' ] := crbracket;
	ccclass [ '[' ] := clbracket;   ccclass [ ']' ] := crbracket;
	ccclass [ '{' ] := clbracket;   ccclass [ '}' ] := crbracket;  { CURLY }
	ccclass [ ',' ] := csolochar;
	ccclass [ ';' ] := csolochar;   ccclass [ '!' ] := csolochar;
	ccclass [ '_' ] := cunderscore;
	ccclass [ '+' ] := csign;   ccclass [ '-' ] := csign;
	ccclass [ '*' ] := csymch;   ccclass [ '/' ] := csymch;
	ccclass [ '=' ] := csymch;   ccclass [ '@' ] := csymch;
	ccclass [ '#' ] := csymch;   ccclass [ '$' ] := csymch;
	ccclass [ '&' ] := csymch;   ccclass [ '.' ] := cperiod;
	ccclass [ ':' ] := csymch;   ccclass [ '?' ] := csymch;
	ccclass [ '"' ] := cstringquote;   ccclass [ '''' ] := cidquote;
	ccclass [ '%' ] := cpercent;   ccclass [ '|' ] := cbar;
							   *****************)
	readln ( mc );                                 { load opop table }
	for op:= fx to xf do begin
	    for op2:= fx to xf do begin
		read ( mc, c );
		if (c < conflict) or (c > redifnoarg) then errorr ( badkernel );
		opop[ op, op2 ] :=  c
	    end;
	    readln ( mc );
	end;
	readln ( mc );
	for lex1:= idarglex to barlex do begin         { load eqpr table }
	    for lex2:= barlex to rroundlex do begin
		read ( mc, c );
		if (c < badmatch) or (c > closecurly) then errorr ( badkernel );
		eqpr[ lex1, lex2 ] :=  c
	    end;
	    readln ( mc );
	end;
	for a:= atmsemcol to atmuser do std [ a ] :=  rdatom;
	lastasserted:= std [atmtag];  {cf. "redefine"}
	for op:= fx to xf do opatoms[op]:= rdatom;
	for s:= idfail to idtrue do begin
		new ( p );
		with p^ do begin
			next:= nil;   nvars:= 0;
			systemroutine:= true;   sysid:= s
		end;
		at [ rdatom ]. proc:= p
	end {for};
	{ create nil, character & standard integer prototypes }
	mt [ nilprotx ] :=  std [atmnil] ;
	for k:= 0 to maxchar do begin
		buildname ( chr ( k ) );
		mt [ minchprotx + k ] :=  findatom ( wrapname, 0 )
	end;
	for k:= minstdint to maxstdint do mkint ( k );
	{ create error call sequence & user file name prot }
	prot:= initprot ( std [atmerror] );   newpvararg ( prot , 1, 0 );
	prot:= wrapprot ( prot );
	errorcallseq:= initprot ( std [atmdot] );
	newparg(errorcallseq,1,prot );   newparg(errorcallseq, 2, nilprotx);
	errorcallseq:= wrapprot ( errorcallseq );
	userprot:= wrapprot ( initprot ( std [atmuser] ) );
	seeobj:= userprot;   tellobj:= userprot;
	{ locate names of ',' & '|', create std op decl for ',' }
	commaopname:= mt[charprot(',')];
	baropname:= mt[ charprot ( '|' ) ];
	new ( pop );   pop^.name:= commaopname;
	pop^.priority:= commapriority;  pop^.kinds:= [xfy];  storeop ( pop );
end {loadsyskernel};



begin {********** toy prolog ***********}
	initvars;
	loadsyskernel;
2:      repeat  doread ( false, dummy );  { this fills goalstmnt & ngoalvars }
		resolve ( goalstmnt, ngoalvars )
	until terminate;
1:      closefile ( true );   closefile ( false )
end.
