#!/bin/bash
#
#                               mutation tests
#
# usage:
#
#               ./test.sh <files>
#
# where <files> is a list of files to mutate
# if <files> is not present, use the default file model.pas
# Try to crash the compiler by feeding it mutated versions of a model file (or files)
#
# This script provides the necessary inputs to the mutate program and scans
# the output for any compiler faultss
#
#

dotests() {
    for MODEL in $FILES; do
        printf "\rmutation tests with $MODEL ...       "
        for A in remove replace add ; do
            echo >> mlog-$A
            echo testing with $MODEL >> mlog-$A
            ./mutate-$A ../$MODEL ${MYTEMP}$A.pas ${MYTEMP}$A.lst >> mlog-$A 2>&1   &
        done
        wait
    done
}

if [ -z $P5CDIR ] ; then
    echo "\$P5CDIR is not set, bailing out"
    exit 1;
fi

echo
echo "mutation tests for remove, replace & add tokens"

if [ -z $1 ]; then
    if ! [ -f model.pas ]; then
        echo "test file model.pas does not exist - quitting"
        exit
    fi
    FILES=model.pas
    echo "mutant test using $FILES"
else
    FILES=$@
fi


DIR=$(mktemp -d mut-XXX) || { echo "Failed to create work directory"; exit 1; }

# in msys /tmp is a virtual dir and is not visible to some programs
if ls / | grep -q tmp; then
    MYTEMP=/tmp/${DIR}/
    mkdir $MYTEMP
else
    MYTEMP=
    echo  "no /tmp dir found, using $DIR"
fi

cp mutate.pas ${DIR}
cd $DIR
echo "working in $(pwd)"

for A in remove replace add ; do
    echo rebuilding mutate program for $A
    $P5CDIR/pc -O3 -DACTION=$A \
               -DCOMMAND=\'$P5CDIR'/p5x '${MYTEMP}$A'.pas '${MYTEMP}$A'.c > '${MYTEMP}$A'.lst'\' mutate.pas
    if [ ! -x mutate ]; then
        echo rebuild mutate program not successful for $A - quitting
        exit
    fi
    mv mutate mutate-$A
done

echo

if test -n "$BASH_VERSION" ; then
    time dotests
else
    dotests
fi

echo
if grep --before-context=1 fault mlog-* ; then
    read -rsp $'Press any key' -n1
   echo
else
    echo "no faults found"
fi

if grep --before-context=1 fatal mlog-* ; then
   read -rsp $'Press any key' -n1
   echo
else
    echo "no fatals found"

    echo
    if [ -z $1 ]; then
        read -p 'Press RETURN to continue ... ' REPLY
        echo
    fi
fi

#rmdir /tmp/$DIR
